# This file is part of thermotools.
#
# Copyright 2015, 2016 Computational Molecular Biology Group, Freie Universitaet Berlin (GER)
#
# thermotools is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

r"""
Python interface to the WHAM estimator's lowlevel functions.
"""

import numpy as _np
cimport numpy as _np

from warnings import warn as _warn
from msmtools.util.exceptions import NotConvergedWarning as _NotConvergedWarning

from .callback import CallbackInterrupt

__all__ = [
    'update_conf_energies',
    'update_therm_energies',
    'normalize',
    'get_loglikelihood',
    'estimate']

cdef extern from "_wham.h":
    void _wham_update_conf_energies(
        double *log_therm_state_counts, double *log_conf_state_counts,
        double *therm_energies, double *bias_energies,
        int n_therm_states, int n_conf_states, double *scratch_T, double *conf_energies)
    void _wham_update_therm_energies(
        double *conf_energies, double *bias_energies, int n_therm_states, int n_conf_states,
        double *scratch_M, double *therm_energies)
    void _wham_normalize(
        int n_therm_states, int n_conf_states,
        double *scratch_M, double *therm_energies, double *conf_energies)
    double _wham_get_loglikelihood(
        int *therm_state_counts, int *conf_state_counts,
        double *therm_energies, double *conf_energies,
        int n_therm_states, int n_conf_states, double *scratch_S)

def update_conf_energies(
    _np.ndarray[double, ndim=1, mode="c"] log_therm_state_counts not None,
    _np.ndarray[double, ndim=1, mode="c"] log_conf_state_counts not None,
    _np.ndarray[double, ndim=1, mode="c"] therm_energies not None,
    _np.ndarray[double, ndim=2, mode="c"] bias_energies not None,
    _np.ndarray[double, ndim=1, mode="c"] scratch_T not None,
    _np.ndarray[double, ndim=1, mode="c"] conf_energies not None):
    r"""
    Calculate the reduced unbiased free energies.
        
    Parameters
    ----------
    log_therm_state_counts : numpy.ndarray(shape=(T), dtype=numpy.float64)
        logarithm of the state counts in each of the T thermodynamic states
    log_conf_state_counts : numpy.ndarray(shape=(M), dtype=numpy.float64)
        logarithm of the state counts in each of the M configurational states
    therm_energies : numpy.ndarray(shape=(T), dtype=numpy.float64)
        reduced free energies of the T thermodynamic states
    bias_energies : numpy.ndarray(shape=(T, M), dtype=numpy.intc)
        reduced bias energies of the T thermodynamic and M configurational states
    scratch_T : numpy.ndarray(shape=(T), dtype=numpy.float64)
        scratch array
    conf_energies : numpy.ndarray(shape=(M,), dtype=numpy.float64)
        reduced unbiased configurational energies

    Notes
    -----
    The update_conf_energies() function computes

    .. math ::
        \text{conf_energies}_i = -\ln\left( \frac{
                \sum_{K=0}^{N_T-1} N_i^{(K)}
            }{
                \sum_{K=0}^{N_T-1} \exp(f^{(K)}-b_i^{(K)}) \sum_{j=0}^{N_M-1} N_j^{(K)}
            }\right)

    which equals to

    .. math ::
        \text{conf_energies}_i = \ln\left(
                \sum_{K=0}^{N_T-1} \exp\left(
                    f^{(K)} - b_i^{(K)} + \ln\left( \sum_{j=0}^{N_M-1} N_j^{(K)} \right)
                \right)
            \right) - \ln\left( \sum_{K=0}^{N_T-1} N_i^{(K)} \right)

    in the logsumexp scheme. Afterwards, we apply

    .. math ::
        \text{conf_energies} \leftarrow  \text{conf_energies} - \min(\text{conf_energies})

    """
    _wham_update_conf_energies(
        <double*> _np.PyArray_DATA(log_therm_state_counts),
        <double*> _np.PyArray_DATA(log_conf_state_counts),
        <double*> _np.PyArray_DATA(therm_energies),
        <double*> _np.PyArray_DATA(bias_energies),
        bias_energies.shape[0],
        bias_energies.shape[1],
        <double*> _np.PyArray_DATA(scratch_T),
        <double*> _np.PyArray_DATA(conf_energies))

def update_therm_energies(
    _np.ndarray[double, ndim=1, mode="c"] conf_energies not None,
    _np.ndarray[double, ndim=2, mode="c"] bias_energies not None,
    _np.ndarray[double, ndim=1, mode="c"] scratch_M not None,
    _np.ndarray[double, ndim=1, mode="c"] therm_energies not None):
    r"""
    Calculate the reduced thermodynamic free energies therm_energies
        
    Parameters
    ----------
    conf_energies : numpy.ndarray(shape=(M,), dtype=numpy.float64)
        reduced unbiased configurational energies
    bias_energies : numpy.ndarray(shape=(T, M), dtype=numpy.intc)
        reduced bias energies of the T thermodynamic and M configurational states
    scratch_M : numpy.ndarray(shape=(M), dtype=numpy.float64)
        scratch array
    therm_energies : numpy.ndarray(shape=(T), dtype=numpy.float64)
        target array for the reduced free energies of the T thermodynamic states

    Notes
    -----
    The update_therm_energies() function computes

    .. math ::
        f^{(K)} = -\ln\left(
                \sum_{i=0}^{N_M-1} \exp(-b_i^{(K)}-\text{conf_energies}_i)
            \right)

    which is already in the logsumexp form.
    """
    _wham_update_therm_energies(
        <double*> _np.PyArray_DATA(conf_energies),
        <double*> _np.PyArray_DATA(bias_energies),
        bias_energies.shape[0],
        bias_energies.shape[1],
        <double*> _np.PyArray_DATA(scratch_M),
        <double*> _np.PyArray_DATA(therm_energies))

def normalize(
    _np.ndarray[double, ndim=1, mode="c"] scratch_M not None,
    _np.ndarray[double, ndim=1, mode="c"] therm_energies not None,
    _np.ndarray[double, ndim=1, mode="c"] conf_energies not None):
    r"""
    Normalize the unbiased reduced free energies and shift the thermodynamic
    free energies accordingly
        
    Parameters
    ----------
    scratch_M : numpy.ndarray(shape=(M), dtype=numpy.float64)
        scratch array
    therm_energies : numpy.ndarray(shape=(T), dtype=numpy.float64)
        reduced free energies of the T thermodynamic states
    conf_energies : numpy.ndarray(shape=(M,), dtype=numpy.float64)
        reduced unbiased configurational energies
    """
    _wham_normalize(
        therm_energies.shape[0],
        conf_energies.shape[0],
        <double*> _np.PyArray_DATA(scratch_M),
        <double*> _np.PyArray_DATA(therm_energies),
        <double*> _np.PyArray_DATA(conf_energies))

def get_loglikelihood(
    _np.ndarray[int, ndim=1, mode="c"] therm_state_counts not None,
    _np.ndarray[int, ndim=1, mode="c"] conf_state_counts not None,
    _np.ndarray[double, ndim=1, mode="c"] therm_energies not None,
    _np.ndarray[double, ndim=1, mode="c"] conf_energies not None,
    _np.ndarray[double, ndim=1, mode="c"] scratch_S not None):
    r"""
    Compute the loglikelihood of the estimated reduced free energies.

    Parameters
    ----------


    Returns
    -------
    loglikelihood : float
        loglikelihood of the reduced free energies given the observed state counts
    """
    return _wham_get_loglikelihood(
        <int*> _np.PyArray_DATA(therm_state_counts),
        <int*> _np.PyArray_DATA(conf_state_counts),
        <double*> _np.PyArray_DATA(therm_energies),
        <double*> _np.PyArray_DATA(conf_energies),
        therm_state_counts.shape[0],
        conf_state_counts.shape[0],
        <double*> _np.PyArray_DATA(scratch_S))

def estimate(
    state_counts, bias_energies,
    maxiter=1000, maxerr=1.0E-8,
    therm_energies=None, conf_energies=None,
    save_convergence_info=0, callback=None):
    r"""
    Estimate the unbiased reduced free energies and thermodynamic free energies
        
    Parameters
    ----------
    state_counts : numpy.ndarray(shape=(T, M), dtype=numpy.intc)
        state counts in the T thermodynamic and M configurational states
    bias_energies : numpy.ndarray(shape=(T, M), dtype=numpy.intc)
        reduced bias energies of the T thermodynamic and M configurational states
    maxiter : int
        maximum number of iterations
    maxerr : float
        convergence criterion based on absolute change in free energies
    therm_energies : numpy.ndarray(shape=(T,), dtype=numpy.float64), OPTIONAL
        initial guess for the reduced thermodynamic energies
    conf_energies : numpy.ndarray(shape=(M,), dtype=numpy.float64)
        initial guess for the reduced unbiased free energies
    save_convergence_info : int, optional
        every save_convergence_info iteration steps, store the actual increment
        and the actual loglikelihood

    Returns
    -------
    therm_energies : numpy.ndarray(shape=(T,), dtype=numpy.float64)
        reduced thermodynamic energies
    conf_energies : numpy.ndarray(shape=(M,), dtype=numpy.float64)
        reduced unbiased configurational energies
    increments : numpy.ndarray(dtype=numpy.float64, ndim=1)
        stored sequence of increments
    loglikelihoods : numpy.ndarray(dtype=numpy.float64, ndim=1)
        stored sequence of loglikelihoods

    Notes
    -----
    This function calls the previously defined update functions to estimate the reduced
    configuration energies of the unbiased thermodynamic state and the reduced thermodynamic
    energies by means of a fixed point iteration.
    """
    T = state_counts.shape[0]
    M = state_counts.shape[1]
    S = T + M
    therm_state_counts = state_counts.sum(axis=1).astype(_np.intc)
    conf_state_counts = state_counts.sum(axis=0).astype(_np.intc)
    log_therm_state_counts = _np.log(therm_state_counts).astype(_np.float64)
    log_conf_state_counts = _np.log(conf_state_counts).astype(_np.float64)
    if therm_energies is None:
        therm_energies = _np.zeros(shape=(T,), dtype=_np.float64)
    if conf_energies is None:
        conf_energies = _np.zeros(shape=(M,), dtype=_np.float64)
    old_therm_energies = therm_energies.copy()
    old_conf_energies = conf_energies.copy()
    scratch = _np.zeros(shape=(S,), dtype=_np.float64)
    increments = []
    loglikelihoods = []
    sci_count = 0
    for m in range(maxiter):
        sci_count += 1
        update_therm_energies(conf_energies, bias_energies, scratch, therm_energies)
        update_conf_energies(
            log_therm_state_counts, log_conf_state_counts, therm_energies, bias_energies,
            scratch, conf_energies)
        delta_therm_energies = _np.abs(therm_energies - old_therm_energies)
        delta_conf_energies = _np.abs(conf_energies - old_conf_energies)
        err = _np.max([_np.max(delta_conf_energies), _np.max(delta_therm_energies)])
        normalize(scratch, therm_energies, conf_energies)
        if sci_count == save_convergence_info:
            sci_count = 0
            increments.append(err)
            loglikelihoods.append(
                get_loglikelihood(
                    therm_state_counts, conf_state_counts, therm_energies, conf_energies, scratch))
        if callback is not None:
            try:
                callback(
                    conf_energies=conf_energies,
                    old_therm_energies=old_therm_energies,
                    old_conf_energies=old_conf_energies,
                    therm_energies=therm_energies,
                    delta_conf_energies=delta_conf_energies,
                    delta_therm_energies=delta_therm_energies,
                    err=err,
                    iteration_step=m,
                    maxiter=maxiter,
                    maxerr=maxerr)
            except CallbackInterrupt:
                break
        if err < maxerr:
            break
        else:
            old_therm_energies[:] = therm_energies[:]
            old_conf_energies[:] = conf_energies[:]
    if err >= maxerr:
        _warn("WHAM did not converge: last increment = %.5e" % err, _NotConvergedWarning)
    if save_convergence_info == 0:
        increments = None
        loglikelihoods = None
    else:
        increments = _np.array(increments, dtype=_np.float64)
        loglikelihoods = _np.array(loglikelihoods, dtype=_np.float64)
    return therm_energies, conf_energies, increments, loglikelihoods



# This file is part of thermotools.
#
# Copyright 2015, 2016 Computational Molecular Biology Group, Freie Universitaet Berlin (GER)
#
# thermotools is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import thermotools.wham as wham
import thermotools.mbar as mbar
import thermotools.dtram as dtram
import numpy as np
from numpy.testing import assert_allclose
from nose.tools import assert_true, assert_raises

from thermotools.callback import CallbackInterrupt, generic_callback_stop

#   ************************************************************************************************
#   test generic_callback_stop
#   ************************************************************************************************

def test_callback_interrupt():
    assert_raises(CallbackInterrupt, generic_callback_stop)
    try:
        generic_callback_stop()
    except CallbackInterrupt as ci:
        assert_true(ci.msg == "STOP")
        assert_true(ci.__str__() == "[CALLBACKINTERRUPT] STOP")

def test_wham_stop():
    T = 5
    M = 10
    therm_energies, conf_energies, increments, loglikelihoods = wham.estimate(
        np.ones(shape=(T, M), dtype=np.intc),
        np.zeros(shape=(T, M), dtype=np.float64),
        maxiter=10, maxerr=-1.0, save_convergence_info=1,
        callback=generic_callback_stop)
    assert_allclose(therm_energies, 0.0, atol=1.0E-15)
    assert_allclose(conf_energies, np.log(M), atol=1.0E-15)
    assert_true(increments.shape[0] == 1)
    assert_true(loglikelihoods.shape[0] == 1)

def test_dtram_stop():
    T = 5
    M = 10
    therm_energies, conf_energies, log_lagrangian_mult, increments, loglikelihoods = dtram.estimate(
        np.ones(shape=(T, M, M), dtype=np.intc),
        np.zeros(shape=(T, M), dtype=np.float64),
        maxiter=10, maxerr=-1.0, save_convergence_info=1,
        callback=generic_callback_stop)
    assert_allclose(therm_energies, 0.0, atol=1.0E-15)
    assert_allclose(conf_energies, np.log(M), atol=1.0E-15)
    assert_allclose(log_lagrangian_mult, np.log(M + dtram.get_prior()), atol=1.0E-15)
    assert_true(increments.shape[0] == 1)
    assert_true(loglikelihoods.shape[0] == 1)



# This file is part of thermotools.
#
# Copyright 2015 Computational Molecular Biology Group, Freie Universitaet Berlin (GER)
#
# thermotools is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import thermotools.mbar as mbar
import numpy as np
from numpy.testing import assert_allclose

ca = np.ascontiguousarray

def test_mbar_fk_with_zeros():
    T = 5
    X = 10
    log_therm_state_counts = np.zeros(shape=(T,), dtype=np.float64)
    bias_energies = np.zeros(shape=(T, X), dtype=np.float64)
    scratch = np.zeros(shape=(T,), dtype=np.float64)
    therm_energies = np.zeros(shape=(T,), dtype=np.float64)
    new_therm_energies = np.zeros(shape=(T,), dtype=np.float64)
    mbar.update_therm_energies(
        log_therm_state_counts, therm_energies, [ca(bias_energies.T)], scratch, new_therm_energies)
    assert_allclose(new_therm_energies, 0.0, atol=1.0E-15)

def test_mbar_fk_with_ascending_bias():
    T = 5
    X = 10
    log_therm_state_counts = np.zeros(shape=(T,), dtype=np.float64)
    bias_energies = np.array([[K]*X for K in range(T)], dtype=np.float64)
    scratch = np.zeros(shape=(T,), dtype=np.float64)
    therm_energies = np.zeros(shape=(T,), dtype=np.float64)
    new_therm_energies = np.zeros(shape=(T,), dtype=np.float64)
    ref = np.array([float(K) - np.log(X) for K in range(T)], dtype=np.float64)
    ref -= ref.min()
    mbar.update_therm_energies(
        log_therm_state_counts, therm_energies, [ca(bias_energies.T)], scratch, new_therm_energies)
    assert_allclose(new_therm_energies, ref, atol=1.0E-15)



# This file is part of thermotools.
#
# Copyright 2015, 2016 Computational Molecular Biology Group, Freie Universitaet Berlin (GER)
#
# thermotools is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

r"""
Python interface to the dTRAM estimator's lowlevel functions.
"""

import numpy as _np
cimport numpy as _np

from warnings import warn as _warn
from msmtools.util.exceptions import NotConvergedWarning as _NotConvergedWarning

from . import util
from .callback import CallbackInterrupt

__all__ = [
    'init_log_lagrangian_mult',
    'update_log_lagrangian_mult',
    'update_conf_energies',
    'estimate_transition_matrices',
    'estimate_transition_matrix',
    'get_therm_energies',
    'normalize',
    'get_loglikelihood',
    'get_prior',
    'get_log_prior',
    'estimate']

cdef extern from "_dtram.h":
    void _dtram_init_log_lagrangian_mult(
        int *count_matrices, int n_therm_states, int n_conf_states, double *log_lagrangian_mult)
    void _dtram_update_log_lagrangian_mult(
        double *log_lagrangian_mult, double *bias_energies, double *conf_energies,
        int *count_matrices, int n_therm_states, int n_conf_states,
        double *scratch_M, double *new_log_lagrangian_mult)
    void _dtram_update_conf_energies(
        double *log_lagrangian_mult, double *bias_energies, double *conf_energies,
        int *count_matrices, int n_therm_states, int n_conf_states,
        double *scratch_TM, double *new_conf_energies)
    void _dtram_estimate_transition_matrix(
        double *log_lagrangian_mult, double *b_i, double *conf_energies, int *count_matrix,
        int n_conf_states, double *scratch_M, double *transition_matrix)
    void _dtram_get_therm_energies(
        double *bias_energies, double *conf_energies, int n_therm_states, int n_conf_states,
        double *scratch_M, double *therm_energies)
    void _dtram_normalize(
        int n_therm_states, int n_conf_states,
        double *scratch_M, double *therm_energies, double *conf_energies)
    double _dtram_get_loglikelihood(
        int *count_matrices, double *transition_matrices,
        int n_therm_states, int n_conf_states)
    double _dtram_get_prior()
    double _dtram_get_log_prior()

def init_log_lagrangian_mult(
    _np.ndarray[int, ndim=3, mode="c"] count_matrices not None):
    r"""
    Set the logarithm of the Lagrangian multipliers with an initial guess based
    on the transition counts.

    Parameters
    ----------
    count_matrices : numpy.ndarray(shape=(T, M, M), dtype=numpy.intc)
        multistate count matrix

    Returns
    -------
    log_lagrangian_mult : numpy.ndarray(shape=(T, M), dtype=numpy.float64)
        logarithm of the Lagrangian multipliers
    """
    log_lagrangian_mult = _np.zeros(
        shape=(count_matrices.shape[0], count_matrices.shape[1]), dtype=_np.float64)
    _dtram_init_log_lagrangian_mult(
        <int*> _np.PyArray_DATA(count_matrices),
        log_lagrangian_mult.shape[0],
        log_lagrangian_mult.shape[1],
        <double*> _np.PyArray_DATA(log_lagrangian_mult))
    return log_lagrangian_mult

def update_log_lagrangian_mult(
    _np.ndarray[double, ndim=2, mode="c"] log_lagrangian_mult not None,
    _np.ndarray[double, ndim=2, mode="c"] bias_energies not None,
    _np.ndarray[double, ndim=1, mode="c"] conf_energies not None,
    _np.ndarray[int, ndim=3, mode="c"] count_matrices not None,
    _np.ndarray[double, ndim=1, mode="c"] scratch_M not None,
    _np.ndarray[double, ndim=2, mode="c"] new_log_lagrangian_mult not None):
    r"""
    Update the logarithms of the Lagrangian multipliers

    Parameters
    ----------
    log_lagrangian_mult : numpy.ndarray(shape=(T, M), dtype=numpy.float64)
        log of the Lagrangian multipliers
    bias_energies : numpy.ndarray(shape=(T, M), dtype=numpy.intc)
        reduced bias energies of the T thermodynamic and M configurational states
    conf_energies : numpy.ndarray(shape=(M,), dtype=numpy.float64)
        reduced unbiased configurational energies
    count_matrices : numpy.ndarray(shape=(T, M, M), dtype=numpy.intc)
        multistate count matrix
    scratch_M : numpy.ndarray(shape=(M,), dtype=numpy.float64)
        scratch array for logsumexp operations
    new_log_lagrangian_mult : numpy.ndarray(shape=(T, M), dtype=numpy.float64)
        target array for the logarithm of the Lagrangian multipliers
    """
    _dtram_update_log_lagrangian_mult(
        <double*> _np.PyArray_DATA(log_lagrangian_mult),
        <double*> _np.PyArray_DATA(bias_energies),
        <double*> _np.PyArray_DATA(conf_energies),
        <int*> _np.PyArray_DATA(count_matrices),
        log_lagrangian_mult.shape[0],
        log_lagrangian_mult.shape[1],
        <double*> _np.PyArray_DATA(scratch_M),
        <double*> _np.PyArray_DATA(new_log_lagrangian_mult))

def update_conf_energies(
    _np.ndarray[double, ndim=2, mode="c"] log_lagrangian_mult not None,
    _np.ndarray[double, ndim=2, mode="c"] bias_energies not None,
    _np.ndarray[double, ndim=1, mode="c"] conf_energies not None,
    _np.ndarray[int, ndim=3, mode="c"] count_matrices not None,
    _np.ndarray[double, ndim=2, mode="c"] scratch_TM not None,
    _np.ndarray[double, ndim=1, mode="c"] new_conf_energies not None):
    r"""
    Update the reduced unbiased free energies.

    Parameters
    ----------
    log_lagrangian_mult : numpy.ndarray(shape=(T, M), dtype=numpy.float64)
        logarithm of the Lagrangian multipliers
    bias_energies : numpy.ndarray(shape=(T, M), dtype=numpy.intc)
        reduced bias energies of the T thermodynamic and M configurational states
    conf_energies : numpy.ndarray(shape=(M,), dtype=numpy.float64)
        reduced unbiased configurational energies
    count_matrices : numpy.ndarray(shape=(T, M, M), dtype=numpy.intc)
        multistate count matrix
    scratch_TM : numpy.ndarray(shape=(T, M), dtype=numpy.float64)
        scratch array for logsumexp operations
    new_conf_energies : numpy.ndarray(shape=(M,), dtype=numpy.float64)
        target array for the reduced unbiased configurational energies
    """
    _dtram_update_conf_energies(
        <double*> _np.PyArray_DATA(log_lagrangian_mult),
        <double*> _np.PyArray_DATA(bias_energies),
        <double*> _np.PyArray_DATA(conf_energies),
        <int*> _np.PyArray_DATA(count_matrices),
        log_lagrangian_mult.shape[0],
        log_lagrangian_mult.shape[1],
        <double*> _np.PyArray_DATA(scratch_TM),
        <double*> _np.PyArray_DATA(new_conf_energies))

def estimate_transition_matrices(
    _np.ndarray[double, ndim=2, mode="c"] log_lagrangian_mult not None,
    _np.ndarray[double, ndim=2, mode="c"] bias_energies not None,
    _np.ndarray[double, ndim=1, mode="c"] conf_energies not None,
    _np.ndarray[int, ndim=3, mode="c"] count_matrices not None,
    _np.ndarray[double, ndim=1, mode="c"] scratch_M not None):
    r"""
    Compute the transition matrices for all thermodynamic states.

    Parameters
    ----------
    log_lagrangian_mult : numpy.ndarray(shape=(T, M), dtype=numpy.float64)
        logarithm of the Lagrangian multipliers
    bias_energies : numpy.ndarray(shape=(T, M), dtype=numpy.intc)
        reduced bias energies of the T thermodynamic and M configurational states
    conf_energies : numpy.ndarray(shape=(M,), dtype=numpy.float64)
        reduced unbiased configurational energies
    count_matrices : numpy.ndarray(shape=(T, M, M), dtype=numpy.intc)
        multistate count matrix
    scratch_M : numpy.ndarray(shape=(M,), dtype=numpy.float64)
        scratch array for logsumexp operations

    Returns
    -------
    transition_matrices : numpy.ndarray(shape=(T, M, M), dtype=numpy.float64)
        multistate transition matrix
    """
    transition_matrices = _np.zeros(
        shape=(count_matrices.shape[0], count_matrices.shape[1], count_matrices.shape[2]),
        dtype=_np.float64)
    for K in range(log_lagrangian_mult.shape[0]):
        transition_matrices[K, :, :] = estimate_transition_matrix(
            log_lagrangian_mult, bias_energies, conf_energies, count_matrices, scratch_M, K)[:, :]
    return transition_matrices

def estimate_transition_matrix(
    _np.ndarray[double, ndim=2, mode="c"] log_lagrangian_mult not None,
    _np.ndarray[double, ndim=2, mode="c"] bias_energies not None,
    _np.ndarray[double, ndim=1, mode="c"] conf_energies not None,
    _np.ndarray[int, ndim=3, mode="c"] count_matrices not None,
    _np.ndarray[double, ndim=1, mode="c"] scratch_M not None,
    therm_state):
    r"""
    Compute the transition matrix for a single thermodynamic state.

    Parameters
    ----------
    log_lagrangian_mult : numpy.ndarray(shape=(T, M), dtype=numpy.float64)
        log of the Lagrangian multipliers
    bias_energies : numpy.ndarray(shape=(T, M), dtype=numpy.intc)
        reduced bias energies of the T thermodynamic and M configurational states
    conf_energies : numpy.ndarray(shape=(M,), dtype=numpy.float64)
        reduced unbiased configurational energies
    count_matrices : numpy.ndarray(shape=(T, M, M), dtype=numpy.intc)
        multistate count matrix
    scratch_M : numpy.ndarray(shape=(M,), dtype=numpy.float64)
        scratch array for logsumexp operations
    therm_state : int
        target thermodynamic state

    Returns
    -------
    transition_matrix : numpy.ndarray(shape=(M, M), dtype=numpy.float64)
        transition matrix for the target thermodynamic state
    """
    transition_matrix = _np.zeros(
        shape=(conf_energies.shape[0], conf_energies.shape[0]), dtype=_np.float64)
    _dtram_estimate_transition_matrix(
        <double*> _np.PyArray_DATA(_np.ascontiguousarray(log_lagrangian_mult[therm_state, :])),
        <double*> _np.PyArray_DATA(_np.ascontiguousarray(bias_energies[therm_state, :])),
        <double*> _np.PyArray_DATA(conf_energies),
        <int*> _np.PyArray_DATA(_np.ascontiguousarray(count_matrices[therm_state, :, :])),
        conf_energies.shape[0],
        <double*> _np.PyArray_DATA(scratch_M),
        <double*> _np.PyArray_DATA(transition_matrix))
    return transition_matrix

def get_therm_energies(
    _np.ndarray[double, ndim=2, mode="c"] bias_energies not None,
    _np.ndarray[double, ndim=1, mode="c"] conf_energies not None,
    _np.ndarray[double, ndim=1, mode="c"] scratch_M not None,
    _np.ndarray[double, ndim=1, mode="c"] therm_energies=None):
    r"""
    Compute the reduced thermodynamic energies.

    Parameters
    ----------
    bias_energies : numpy.ndarray(shape=(T, M), dtype=numpy.intc)
        reduced bias energies of the T thermodynamic and M markov states
    conf_energies : numpy.ndarray(shape=(M,), dtype=numpy.float64)
        reduced unbiased configurational energies
    scratch_M : numpy.ndarray(shape=(M,), dtype=numpy.float64)
        scratch array for logsumexp operations
    therm_energies : numpy.ndarray(shape=(T,), dtype=numpy.float64), optional
        target array for the reduced thermodynamic energies

    Returns
    -------
    therm_energies : numpy.ndarray(shape=(T,), dtype=numpy.float64)
        reduced thermodynamic energies
    """
    if therm_energies is None:
        therm_energies = _np.zeros(shape=(bias_energies.shape[0],), dtype=_np.float64)
    _dtram_get_therm_energies(
        <double*> _np.PyArray_DATA(bias_energies),
        <double*> _np.PyArray_DATA(conf_energies),
        bias_energies.shape[0],
        bias_energies.shape[1],
        <double*> _np.PyArray_DATA(scratch_M),
        <double*> _np.PyArray_DATA(therm_energies))
    return therm_energies

def normalize(
    _np.ndarray[double, ndim=1, mode="c"] scratch_M not None,
    _np.ndarray[double, ndim=1, mode="c"] therm_energies not None,
    _np.ndarray[double, ndim=1, mode="c"] conf_energies not None):
    r"""
    Normalize the reduced unbiased configurational energies and shift the reduced thermodynamic
    energies accordingly.

    Parameters
    ----------
    scratch_M : numpy.ndarray(shape=(M), dtype=numpy.float64)
        scratch array for logsumexp operations
    therm_energies : numpy.ndarray(shape=(T,), dtype=numpy.intc)
        reduced thermodynamic energies
    conf_energies : numpy.ndarray(shape=(M,), dtype=numpy.float64)
        reduced unbiased configurational energies
    """
    _dtram_normalize(
        therm_energies.shape[0],
        conf_energies.shape[0],
        <double*> _np.PyArray_DATA(scratch_M),
        <double*> _np.PyArray_DATA(therm_energies),
        <double*> _np.PyArray_DATA(conf_energies))

def get_loglikelihood(
    _np.ndarray[int, ndim=3, mode="c"] count_matrices not None,
    _np.ndarray[double, ndim=3, mode="c"] transition_matrices not None):
    r"""
    Compute the loglikelihood of the estimated transition matrices.

    Parameters
    ----------
    count_matrices : numpy.ndarray(shape=(T, M, M), dtype=numpy.intc)
        multistate count matrix
    transition_matrices : numpy.ndarray(shape=(T, M, M), dtype=numpy.float64)
        multistate transition matrix

    Returns
    -------
    loglikelihood : float
        loglikelihood of the multistate transition matrix given the observed multistate count matrix
    """
    return _dtram_get_loglikelihood(
        <int*> _np.PyArray_DATA(count_matrices),
        <double*> _np.PyArray_DATA(transition_matrices),
        count_matrices.shape[0],
        count_matrices.shape[1])

def get_prior():
    r"""
    Returns
    -------
    prior : float
        the internally used prior on the diagonal transition counts
    """
    return _dtram_get_prior()

def get_log_prior():
    r"""
    Returns
    -------
    log_prior : float
        the internally used log(prior) on the diagonal transition counts
    """
    return _dtram_get_log_prior()

def estimate(
    count_matrices, bias_energies,
    maxiter=1000, maxerr=1.0E-8,
    log_lagrangian_mult=None, conf_energies=None,
    save_convergence_info=0, callback=None):
    r"""
    Estimate the reduced unbiased and thermodynamic free energies.
        
    Parameters
    ----------
    count_matrices : numpy.ndarray(shape=(T, M, M), dtype=numpy.intc)
        multistate count matrix
    bias_energies : numpy.ndarray(shape=(T, M), dtype=numpy.float64)
        reduced bias energies in the T thermodynamic and M configurational states
    maxiter : int
        maximum number of iterations
    maxerr : float
        convergence criterion based on absolute change in the rediced free energies
    conf_energies : numpy.ndarray(shape=(M,), dtype=numpy.float64)
        initial guess for the reduced unbiased free energies
    log_lagrangian_mult : numpy.ndarray(shape=(T, M), dtype=numpy.float64), optional
        initial guess for the logarithm of the Lagrangian multipliers
    save_convergence_info : int, optional
        every save_convergence_info iteration steps, store the actual increment
        and the actual loglikelihood

    Returns
    -------
    therm_energies : numpy.ndarray(shape=(T,), dtype=numpy.float64)
        reduced thermodynamic energies
    conf_energies : numpy.ndarray(shape=(M,), dtype=numpy.float64)
        reduced unbiased configurational energies
    log_lagrangian_mult : numpy.ndarray(shape=(T, M), dtype=numpy.float64)
        logarithm of the Lagrangian multipliers
    increments : numpy.ndarray(dtype=numpy.float64, ndim=1)
        stored sequence of increments
    loglikelihoods : numpy.ndarray(dtype=numpy.float64, ndim=1)
        stored sequence of loglikelihoods

    Notes
    -----
    This function calls the previously defined update functions to estimate the reduced
    configuration energies of the unbiased thermodynamic state, the reduced thermodynamic
    energies, and the logarithms of the Lagarangian multipliers by means of a fixed point
    iteration.
    """
    if log_lagrangian_mult is None:
        log_lagrangian_mult = init_log_lagrangian_mult(count_matrices)
    if conf_energies is None:
        conf_energies = _np.zeros(shape=bias_energies.shape[1], dtype=_np.float64)
    increments = []
    loglikelihoods = []
    sci_count = 0
    scratch_TM = _np.zeros(shape=bias_energies.shape, dtype=_np.float64)
    scratch_M = _np.zeros(shape=conf_energies.shape, dtype=_np.float64)
    therm_energies = _np.zeros(shape=(bias_energies.shape[0],), dtype=_np.float64)
    old_log_lagrangian_mult = log_lagrangian_mult.copy()
    old_conf_energies = conf_energies.copy()
    old_therm_energies = therm_energies.copy()
    for m in range(maxiter):
        sci_count += 1
        update_log_lagrangian_mult(
            old_log_lagrangian_mult, bias_energies, conf_energies, count_matrices,
            scratch_M, log_lagrangian_mult)
        update_conf_energies(
            log_lagrangian_mult, bias_energies, old_conf_energies, count_matrices,
            scratch_TM, conf_energies)
        therm_energies = get_therm_energies(
            bias_energies, conf_energies, scratch_M, therm_energies=therm_energies)
        delta_conf_energies = _np.abs((conf_energies - old_conf_energies))
        delta_therm_energies = _np.abs((therm_energies - old_therm_energies))
        normalize(scratch_M, therm_energies, conf_energies)
        err = _np.max([_np.max(delta_conf_energies), _np.max(delta_therm_energies)])
        if sci_count == save_convergence_info:
            sci_count = 0
            increments.append(err)
            loglikelihoods.append(get_loglikelihood(count_matrices, estimate_transition_matrices(
                log_lagrangian_mult, bias_energies, conf_energies, count_matrices, scratch_M)))
        if callback is not None:
            try:
                callback(
                    log_lagrangian_mult=log_lagrangian_mult,
                    conf_energies=conf_energies,
                    old_therm_energies=old_therm_energies,
                    old_log_lagrangian_mult=old_log_lagrangian_mult,
                    old_conf_energies=old_conf_energies,
                    therm_energies=therm_energies,
                    delta_conf_energies=delta_conf_energies,
                    delta_therm_energies=delta_therm_energies,
                    err=err,
                    iteration_step=m,
                    maxiter=maxiter,
                    maxerr=maxerr)
            except CallbackInterrupt:
                break
        if err < maxerr:
            break
        else:
            old_log_lagrangian_mult[:] = log_lagrangian_mult[:]
            old_conf_energies[:] = conf_energies[:]
            old_therm_energies[:] = therm_energies[:]
    if err >= maxerr:
        _warn("dTRAM did not converge: last increment = %.5e" % err, _NotConvergedWarning)
    if save_convergence_info == 0:
        increments = None
        loglikelihoods = None
    else:
        increments = _np.array(increments, dtype=_np.float64)
        loglikelihoods = _np.array(loglikelihoods, dtype=_np.float64)
    return therm_energies, conf_energies, log_lagrangian_mult, increments, loglikelihoods



# This file is part of thermotools.
#
# Copyright 2015-2017 Computational Molecular Biology Group, Freie Universitaet Berlin (GER)
#
# thermotools is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

r"""
Python interface to utility functions.
"""

cimport cython
import numpy as _np
cimport numpy as _np
from libc.math cimport exp as _libc_exp
from scipy.sparse import csr_matrix as _csr
from msmtools.estimation import count_matrix as _cm

__all__ = [
    'kahan_summation',
    'logsumexp',
    'logsumexp_pair',
    'get_therm_state_break_points',
    'count_matrices',
    'state_counts',
    'restrict_samples_to_cset',
    'get_umbrella_bias',
    'renormalize_transition_matrix',
    'renormalize_transition_matrices']

cdef extern from "_util.h":
    # sorting
    void _mixed_sort(double *array, int L, int R)
    # direct summation schemes
    double _kahan_summation(double *array, int size)
    # logspace summation schemes
    double _logsumexp(double *array, int size, double array_max)
    double _logsumexp_kahan_inplace(double *array, int size, double array_max)
    double _logsumexp_sort_inplace(double *array, int size)
    double _logsumexp_sort_kahan_inplace(double *array, int size)
    double _logsumexp_pair(double a, double b)
    # counting states and transitions
    int _get_therm_state_break_points(int *T_x, int seq_length, int *break_points)
    # bias calculation tools
    void _get_umbrella_bias(
        double *traj, double *umbrella_centers, double *force_constants,
        double *width, double *inverse_width,
        int nsamples, int nthermo, int ndim, double *bias)
    # transition matrix renormalization
    void _renormalize_transition_matrix(double *p, int n_conf_states, double *scratch_M)

####################################################################################################
#   sorting
####################################################################################################

def mixed_sort(_np.ndarray[double, ndim=1, mode="c"] array not None,
    inplace=True):
    r"""
    Sorts the given array using a quicksort/mergesort hybrid.
        
    Parameters
    ----------
    array : numpy.ndarray(dtype=numpy.float64, ndim=1)
        unsorted values
    inplace : boolean
        should the sorting be performed inplace

    Returns
    -------
    sorted_array : numpy.ndarray(dtype=numpy.float64, ndim=1)
        sorted values

    Notes
    -----
    This python wrapper is only to expose the underlying C function to the nose test suite.
    """
    x = array
    if not inplace:
        x = array.copy()
    _mixed_sort(<double*> _np.PyArray_DATA(x), 0, x.shape[0] - 1)
    return x

####################################################################################################
#   direct summation schemes
####################################################################################################

def kahan_summation(_np.ndarray[double, ndim=1, mode="c"] array not None,
    sort_array=True,
    inplace=True):
    r"""
    Sums the array using Kahan's algorithm.
        
    Parameters
    ----------
    array : numpy.ndarray(dtype=numpy.float64, ndim=1)
        (unsorted) values
    sort_array : boolean
        should the array be sorted before summation
    inplace : boolean
        should the sorting be performed inplace

    Returns
    -------
    sum : float
        sum of the array's values
    """
    x = array
    if sort_array:
        x = mixed_sort(x, inplace=inplace)
    return _kahan_summation(<double*> _np.PyArray_DATA(x), x.shape[0])

####################################################################################################
#   logspace summation schemes
####################################################################################################

def logsumexp(_np.ndarray[double, ndim=1, mode="c"] array not None,
    sort_array=True,
    inplace=True,
    use_kahan=True):
    r"""
    Perform a summation of an array of exponentials via the logsumexp scheme.
        
    Parameters
    ----------
    array : numpy.ndarray(dtype=numpy.float64, ndim=1)
        arguments of the exponentials
    sort_array : boolean
        should the array be sorted before summation
    inplace : boolean
        should the sorting be performed inplace
    use_kahan : boolean
        use Kahan's algorithm for the actual summation

    Returns
    -------
    ln_sum : float
        logarithm of the sum of exponentials

    Notes
    -----
    The logsumexp() function returns

    .. math ::
        \ln\left( \sum_{i=0}^{n-1} \exp(a_i) \right)

    where the :math:`a_i` are the :math:`n` values in the supplied array.
    """
    x = array
    if not inplace:
        x = array.copy()
    # from now on, we can always use <inplace=True> safely
    if use_kahan:
        if sort_array:
            return _logsumexp_sort_kahan_inplace(<double*> _np.PyArray_DATA(x), x.shape[0])
        else:
            return _logsumexp_kahan_inplace(<double*> _np.PyArray_DATA(x), x.shape[0], x.max())
    else:
        if sort_array:
            return _logsumexp_sort_inplace(<double*> _np.PyArray_DATA(x), x.shape[0])
        else:
            return _logsumexp(<double*> _np.PyArray_DATA(x), x.shape[0], x.max())

def logsumexp_pair(a, b):
    r"""
    Perform a summation of two exponentials via the logsumexp scheme.
        
    Parameters
    ----------
    a : float
        argument of the first exponential
    b : float
        argument of the second exponential

    Returns
    -------
    ln_sum : float
        logarithm of the sum of exponentials

    Notes
    -----
    The logsumexp_pair() function returns

    .. math ::
        \ln\left( \exp(a) + \exp(b) \right)

    where the :math:`a` and :math:`b` are the supplied values.
    """
    return _logsumexp_pair(a, b)

####################################################################################################
#   counting states and transitions
####################################################################################################

def get_therm_state_break_points(
    _np.ndarray[int, ndim=1, mode="c"] T_x not None):
    r"""
    Find thermodynamic state changes within a trajectory.

    Parameters
    ----------
    T_x : numpy.ndarray(shape=(X), dtype=numpy.intc)
        Thermodynamic state sequence of a trajectory of length X

    Returns
    -------
    T_B : numpy.ndarray(shape=(B), dtype=numpy.intc)
        Sequence of first subsequence starting frames
    """
    T_B = _np.zeros(shape=(T_x.shape[0],), dtype=_np.intc)
    nb = _get_therm_state_break_points(
        <int*> _np.PyArray_DATA(T_x),
        T_x.shape[0],
        <int*> _np.PyArray_DATA(T_B))
    return _np.ascontiguousarray(T_B[:nb])

def count_matrices(
    ttrajs, dtrajs, lag, sliding=True, sparse_return=True, nthermo=None, nstates=None):
    # TODO: fix docstring
    r"""
    Count transitions at given lagtime.

    Parameters
    ----------
    dtraj : [numpy.ndarray(shape=(X, 2), dtype=numpy.intc)]
        list of discretized trajectories
    lag : int
        lagtime in trajectory steps
    sliding : bool, optional
        if true the sliding window approach is used for transition counting
    sparse_return : bool (optional)
        whether to return a 3D dense matrix or a list 2D sparse matrices
    nthermo : int, optional
        enforce nthermo count-matrices with shape=(nstates, nstates)
    nstates : int, optional
        enforce count-matrices with shape=(nstates, nstates) for all thermodynamic states

    Returns
    -------
    C_K : [scipy.sparse.coo_matrix] or numpy.ndarray(shape=(T, M, M), dtype=numpy.intc)
        count matrices at given lagtime
    """
    cdef:
        int kmax = int(_np.max([t.max() for t in ttrajs]))
        int nmax = int(_np.max([d.max() for d in dtrajs]))
        int K, i, b
    if nthermo is None:
        nthermo = kmax + 1
    elif nthermo < kmax + 1:
        raise ValueError("nthermo is smaller than the number of observed thermodynamic states")
    if nstates is None:
        nstates = nmax + 1
    elif nstates < nmax + 1:
        raise ValueError("nstates is smaller than the number of observed microstates")
    C_K = [_csr((nstates, nstates), dtype=_np.intc) for _ in range(nthermo)]
    for ttraj, dtraj in zip(ttrajs, dtrajs):
        bp = get_therm_state_break_points(
            _np.require(ttraj, dtype=_np.intc ,requirements=['C', 'A']))
        for b in range(1, bp.shape[0]):
            if bp[b] - bp[b - 1] > lag:
                C_K[ttraj[bp[b - 1]]] = C_K[ttraj[bp[b - 1]]] + _cm(
                    _np.require(dtraj[bp[b - 1]:bp[b]], dtype=_np.intc ,requirements=['C', 'A']),
                    lag, sliding=sliding, sparse_return=True, nstates=nstates)
        if dtraj.shape[0] - bp[-1] > lag:
            C_K[ttraj[bp[-1]]] = C_K[ttraj[bp[-1]]] + _cm(
                _np.require(dtraj[bp[-1]:], dtype=_np.intc ,requirements=['C', 'A']),
                lag, sliding=sliding, sparse_return=True, nstates=nstates)
    if sparse_return:
        return C_K
    return _np.array([C.toarray() for C in C_K], dtype=_np.intc)

def state_counts(ttrajs, dtrajs, nstates=None, nthermo=None):
    # TODO: fix docstring
    r"""
    Count discrete states visits in all thermodynamic states.

    Parameters
    ----------
    dtraj : [numpy.ndarray(shape=(X, 2), dtype=numpy.intc)]
        list of discretized trajectories
    nstates : int, optional
        enforce state count matrix with shape=(nthermo, nstates)
    nthermo : int, optional
        enforce state count matrix with shape=(nthermo, nstates)

    Returns
    -------
    N : numpy.ndarray(shape=(T, M), dtype=numpy.intc)
        state counts
    """
    cdef:
        int kmax = int(_np.max([t.max() for t in ttrajs]))
        int nmax = int(_np.max([d.max() for d in dtrajs]))
        int K, i
    if nthermo is None:
        nthermo = kmax + 1
    elif nthermo < kmax + 1:
        raise ValueError("nthermo is smaller than the number of observed thermodynamic states")
    if nstates is None:
        nstates = nmax + 1
    elif nstates < nmax + 1:
        raise ValueError("nstates is smaller than the number of observed microstates")
    N = _np.zeros(shape=(nthermo, nstates), dtype=_np.intc)
    for d, t in zip(dtrajs, ttrajs):
        for K in range(nthermo):
            idx = (t == K)
            for i in range(nstates):
                N[K, i] += _np.sum(_np.logical_and(idx, (d == i)))
    return N

def restrict_samples_to_cset(state_sequence, bias_energy_sequence, cset):
    r"""
    Restrict full list of samples to a subset and relabel configurational state indices.

    Parameters
    ----------
    state_sequence : numpy.ndarray(shape=(X, 2), dtype=numpy.intc)
        sequence of the thermodynamic and configurational state indices of the X samples
    bias_energy_sequence : numpy.ndarray(shape=(T, X), dtype=numpy.float64)
        sequence of the reduced bias energies for all X samples in all T thermodynamic states
    cset : list
        list of configurational states within the desired set

    Returns
    -------
    new_state_sequence : numpy.ndarray(shape=(Y, 2), dtype=numpy.intc)
        restricted and relabeled sequence of the thermodynamic and configurational state
        indices of the Y valid samples
    new_bias_energy_sequence : numpy.ndarray(shape=(T, Y), dtype=numpy.float64)
        restricted sequence of the reduced bias energies for all Y valid samples in
        all T thermodynamic states
    """
    nmax = int(_np.max([_np.max(state_sequence), _np.max(cset)]))
    mapping = []
    o = 0
    for i in range(nmax + 1):
        if i in cset:
            mapping.append(o)
            o += 1
        else:
            mapping.append(-1)
    mapping = _np.array(mapping, dtype=_np.intc)
    conf_state_sequence = mapping[state_sequence[:, 1]]
    valid_samples = (conf_state_sequence != -1)
    new_state_sequence = _np.ascontiguousarray(state_sequence[valid_samples, :])
    new_state_sequence[:, 1] = conf_state_sequence[valid_samples]
    new_bias_energy_sequence = _np.ascontiguousarray(bias_energy_sequence[:, valid_samples])
    return new_state_sequence, new_bias_energy_sequence

@cython.boundscheck(False)
def _overlap_post_hoc_RE(
    _np.ndarray[double, ndim=2, mode="c"] a not None,
    _np.ndarray[double, ndim=2, mode="c"] b not None,
    factor=1.0):
    cdef:
        unsigned int i, j, n, m
        double n_sum, delta
    n = a.shape[0]
    m = b.shape[0]
    n_sum = 0
    for i in range(n):
        for j in range(m):
            delta = a[i, 0] + b[j, 1] - a[i, 1] - b[j, 0]
            n_sum += min(_libc_exp(delta), 1.0)
    n_avg = n_sum / (n * m)
    return (n + m) * n_avg * factor >= 1.0

####################################################################################################
#   bias calculation tools
####################################################################################################

def get_umbrella_bias(
    _np.ndarray[double, ndim=2, mode="c"] traj not None,
    _np.ndarray[double, ndim=2, mode="c"] umbrella_centers not None,
    _np.ndarray[double, ndim=3, mode="c"] force_constants not None,
    _np.ndarray[double, ndim=1, mode="c"] width not None):
    r"""
    Restrict full list of samples to a subset and relabel configurational state indices.

    Parameters
    ----------
    traj : numpy.ndarray(shape=(X, D), dtype=numpy.float64)
        sequence of the D-dimensional reaction coordinate values of the X samples
    umbrella_centers : numpy.ndarray(shape=(T, D), dtype=numpy.float64)
        sequence of T unique D-dimensional umbrella centers
    force_constants : numpy.ndarray(shape=(T, D, D), dtype=numpy.float64)
        sequence of T unique DxD-dimensional force constants (matrices)

    Returns
    -------
    bias : numpy.ndarray(shape=(X, T), dtype=numpy.float64)
        sequence of the T bias energies for each of the X samples
    """
    nsamples = traj.shape[0]
    nthermo = umbrella_centers.shape[0]
    ndim = traj.shape[1]
    bias = _np.zeros(shape=(nsamples, nthermo), dtype=_np.float64)
    half_width = 0.5 * width
    _get_umbrella_bias(
        <double*> _np.PyArray_DATA(traj),
        <double*> _np.PyArray_DATA(umbrella_centers),
        <double*> _np.PyArray_DATA(force_constants),
        <double*> _np.PyArray_DATA(width),
        <double*> _np.PyArray_DATA(half_width),
        nsamples,
        nthermo,
        ndim,
        <double*> _np.PyArray_DATA(bias))
    return bias

####################################################################################################
#   transition matrix renormalization
####################################################################################################

def renormalize_transition_matrix(
    _np.ndarray[double, ndim=2, mode="c"] P not None,
    _np.ndarray[double, ndim=1, mode="c"] scratch_M not None):
    _renormalize_transition_matrix(
        <double*> _np.PyArray_DATA(P),
        P.shape[0],
        <double*> _np.PyArray_DATA(scratch_M))

def renormalize_transition_matrices(
    _np.ndarray[double, ndim=3, mode="c"] PK not None,
    _np.ndarray[double, ndim=1, mode="c"] scratch_M not None):
    for K in range(PK.shape[0]):
        P = _np.ascontiguousarray(PK[K, :, :])
        _renormalize_transition_matrix(
            <double*> _np.PyArray_DATA(P),
            P.shape[0],
            <double*> _np.PyArray_DATA(scratch_M))
        PK[K, :, :] = P[:, :]



# This file is part of thermotools.
#
# Copyright 2015, 2016 Computational Molecular Biology Group, Freie Universitaet Berlin (GER)
#
# thermotools is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import thermotools.wham as wham
import thermotools.tram as tram
import thermotools.tram_direct as tram_direct
import thermotools.mbar as mbar
import thermotools.mbar_direct as mbar_direct
import thermotools.dtram as dtram
import thermotools.tram as tram
import numpy as np
from numpy.testing import assert_allclose

#   ************************************************************************************************
#   data generation functions
#   ************************************************************************************************

def tower_sample(distribution):
    cdf = np.cumsum(distribution)
    rnd = np.random.rand() * cdf[-1]
    return np.searchsorted(cdf, rnd)

def draw_independent_samples(biased_stationary_distribution, n_samples):
    state_counts = np.zeros(shape=biased_stationary_distribution.shape, dtype=np.intc)
    conf_state_sequence = np.zeros(
        biased_stationary_distribution.shape[0]*n_samples, dtype=np.intc)
    for K in range(biased_stationary_distribution.shape[0]):
        for s in range(n_samples):
            i = tower_sample(biased_stationary_distribution[K, :])
            state_counts[K, i] += 1
            conf_state_sequence[K * n_samples + s] = i
    return conf_state_sequence, state_counts

def draw_transition_counts(transition_matrices, n_samples, x0):
    """generates a discrete Markov chain"""
    count_matrices = np.zeros(shape=transition_matrices.shape, dtype=np.intc)
    conf_state_sequence = np.zeros(
        shape=(transition_matrices.shape[0]*(n_samples+1),), dtype=np.intc)
    state_counts = np.zeros(shape=transition_matrices.shape[0:2], dtype=np.intc)
    h = 0
    for K in range(transition_matrices.shape[0]):
        x = x0
        state_counts[K, x] += 1
        conf_state_sequence[h] = x
        h += 1
        for s in range(n_samples):
            x_new = tower_sample(transition_matrices[K, x, :])
            count_matrices[K, x, x_new] += 1
            x = x_new
            state_counts[K, x] += 1
            conf_state_sequence[h] = x
            h += 1
    return count_matrices, conf_state_sequence, state_counts

#   ************************************************************************************************
#   test class
#   ************************************************************************************************

class TestThreeTwoModel(object):
    @classmethod
    def setup_class(cls):
        cls.energy = np.array([1.0, 2.0, 0.0], dtype=np.float64)
        cls.bias_energies = np.array([[0.0, 0.0, 0.0], 2.0 - cls.energy], dtype=np.float64)
        cls.stationary_distribution = np.exp(-cls.energy) / np.exp(-cls.energy).sum()
        cls.conf_energies = -np.log(cls.stationary_distribution)
        cls.biased_conf_energies = cls.conf_energies[np.newaxis, :] + cls.bias_energies
        cls.conf_partition_function = np.exp(-cls.biased_conf_energies)
        cls.partition_function = cls.conf_partition_function.sum(axis=1)
        cls.therm_energies = -np.log(cls.partition_function)
        cls.biased_stationary_distribution = np.exp(-cls.bias_energies) * \
            cls.stationary_distribution[np.newaxis, :] / cls.partition_function[:, np.newaxis]
        metropolis = cls.energy[np.newaxis, :] - cls.energy[:, np.newaxis]
        metropolis[(metropolis < 0.0)] = 0.0
        selection = np.array([[0.5, 0.5, 0.0], [0.5, 0.0, 0.5], [0.0, 0.5, 0.5]], dtype=np.float64)
        metr_hast = selection * np.exp(-metropolis)
        for i in range(metr_hast.shape[0]):
            metr_hast[i, i] = 0.0
            metr_hast[i, i] = 1.0 - metr_hast[i, :].sum()
        cls.transition_matrices = np.array([metr_hast, selection])
        cls.n_samples = 10000
        cls.conf_state_sequence_ind, cls.state_counts_ind = draw_independent_samples(
            cls.biased_stationary_distribution, cls.n_samples)
        cls.count_matrices, cls.conf_state_sequence, cls.state_counts = draw_transition_counts(
            cls.transition_matrices, cls.n_samples, 0)
    @classmethod
    def teardown_class(cls):
        pass
    def setup(self):
        pass
    def teardown(self):
        pass
    def test_wham(self):
        therm_energies, conf_energies, increments, loglikelihoods = \
            wham.estimate(
                self.state_counts_ind, self.bias_energies, maxiter=50000, maxerr=1.0E-15)
        atol = 1.0E-1
        assert_allclose(therm_energies, self.therm_energies, atol=atol)
        assert_allclose(conf_energies, self.conf_energies, atol=atol)
    def test_mbar(self):
        bias_energy_sequence = np.ascontiguousarray(
            self.bias_energies[:, self.conf_state_sequence_ind].T)
        therm_energies, conf_energies, biased_conf_energies, err_history = mbar.estimate(
            self.state_counts_ind.sum(axis=1),
            [bias_energy_sequence],
            [self.conf_state_sequence_ind],
            maxiter=10000, maxerr=1.0E-15)
        maxerr = 1.0E-1
        assert_allclose(biased_conf_energies, self.biased_conf_energies, atol=maxerr)
        assert_allclose(conf_energies, self.conf_energies, atol=maxerr)
        assert_allclose(therm_energies, self.therm_energies, atol=maxerr)
    def test_mbar_direct(self):
        bias_energy_sequence = np.ascontiguousarray(
            self.bias_energies[:, self.conf_state_sequence_ind].T)
        therm_energies, conf_energies, biased_conf_energies, err_history = mbar_direct.estimate(
            self.state_counts_ind.sum(axis=1),
            [bias_energy_sequence],
            [self.conf_state_sequence_ind],
            maxiter=10000, maxerr=1.0E-15)
        maxerr = 1.0E-1
        assert_allclose(biased_conf_energies, self.biased_conf_energies, atol=maxerr)
        assert_allclose(conf_energies, self.conf_energies, atol=maxerr)
        assert_allclose(therm_energies, self.therm_energies, atol=maxerr)
    def test_dtram(self):
        therm_energies, conf_energies, log_lagrangian_mult, increments, loglikelihoods = \
            dtram.estimate(
                self.count_matrices, self.bias_energies, maxiter=10000, maxerr=1.0E-15)
        transition_matrices = dtram.estimate_transition_matrices(
            log_lagrangian_mult, self.bias_energies, conf_energies, self.count_matrices,
            np.zeros(shape=conf_energies.shape, dtype=np.float64))
        maxerr = 1.0E-1
        assert_allclose(therm_energies, self.therm_energies, atol=maxerr)
        assert_allclose(conf_energies, self.conf_energies, atol=maxerr)
        assert_allclose(transition_matrices, self.transition_matrices, atol=maxerr)
    def test_tram(self):
        bias_energies = np.ascontiguousarray(self.bias_energies[:,self.conf_state_sequence].T)
        biased_conf_energies, conf_energies, therm_energies, log_lagrangian_mult, error_history, logL_history = tram.estimate(
            self.count_matrices, self.state_counts, [bias_energies], [self.conf_state_sequence],
            maxiter=10000, maxerr=1.0E-12, save_convergence_info=10)
        transition_matrices = tram.estimate_transition_matrices(
            log_lagrangian_mult, biased_conf_energies, self.count_matrices, None)
        maxerr = 1.0E-1
        assert_allclose(biased_conf_energies, self.biased_conf_energies, atol=maxerr)
        assert_allclose(conf_energies, self.conf_energies, atol=maxerr)
        assert_allclose(therm_energies, self.therm_energies, atol=maxerr)
        assert_allclose(transition_matrices, self.transition_matrices, atol=maxerr)
        # lower bound on the log-likelihood must be maximal at convergence
        assert np.all(logL_history[-1]+1.E-5>=logL_history[0:-1])
    def test_tram_direct(self):
        bias_energies = np.ascontiguousarray(self.bias_energies[:,self.conf_state_sequence].T)
        biased_conf_energies, conf_energies, therm_energies, log_lagrangian_mult, error_history, logL_history = tram_direct.estimate(
            self.count_matrices, self.state_counts, [bias_energies], [self.conf_state_sequence],
            maxiter=10000, maxerr=1.0E-12, save_convergence_info=10)
        transition_matrices = tram.estimate_transition_matrices(
            log_lagrangian_mult, biased_conf_energies, self.count_matrices, None)
        maxerr = 1.0E-1
        assert_allclose(biased_conf_energies, self.biased_conf_energies, atol=maxerr)
        assert_allclose(conf_energies, self.conf_energies, atol=maxerr)
        assert_allclose(therm_energies, self.therm_energies, atol=maxerr)
        assert_allclose(transition_matrices, self.transition_matrices, atol=maxerr)
        # lower bound on the log-likelihood must be maximal at convergence
        assert np.all(logL_history[-1]+1.E-5>=logL_history[0:-1])        



# This file is part of thermotools.
#
# Copyright 2015, 2016 Computational Molecular Biology Group, Freie Universitaet Berlin (GER)
#
# thermotools is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

r"""
Python interface to the MBAR estimator's lowlevel functions.
"""

import numpy as _np
cimport numpy as _np

from warnings import warn as _warn
from msmtools.util.exceptions import NotConvergedWarning as _NotConvergedWarning

from .callback import CallbackInterrupt

__all__ = [
    'update_therm_energies',
    'get_conf_energies',
    'normalize',
    'get_pointwise_unbiased_free_energies',
    'estimate_therm_energies',
    'estimate']

cdef extern from "_mbar.h":
    void _mbar_update_therm_energies(
        double *log_therm_state_counts, double *therm_energies, double *bias_energy_sequence,
        int n_therm_states, int seq_length, double *scratch_T, double *new_therm_energies)
    void _mbar_get_conf_energies(
        double *log_therm_state_counts, double *therm_energies,
        double *bias_energy_sequence, int * conf_state_sequence,
        int n_therm_states, int n_conf_states, int seq_length,
        double *scratch_T, double *conf_energies, double *biased_conf_energies)
    extern void _mbar_normalize(
        int n_therm_states, int n_conf_states, double *scratch_M,
        double *therm_energies, double *conf_energies, double *biased_conf_energies)
    void _mbar_get_pointwise_unbiased_free_energies(
        int k, double *log_therm_state_counts, double *therm_energies,
        double *bias_energy_sequence,
        int n_therm_states,  int seq_length,
        double *scratch_T, double *pointwise_unbiased_free_energies)


def update_therm_energies(
    _np.ndarray[double, ndim=1, mode="c"] log_therm_state_counts not None,
    _np.ndarray[double, ndim=1, mode="c"] therm_energies not None,
    bias_energy_sequences, # _np.ndarray[double, ndim=2, mode="c"]
    _np.ndarray[double, ndim=1, mode="c"] scratch_T not None,
    _np.ndarray[double, ndim=1, mode="c"] new_therm_energies not None):
    r"""
    Calculate the reduced thermodynamic free energies therm_energies.
        
    Parameters
    ----------
    log_therm_state_counts : numpy.ndarray(shape=(T), dtype=numpy.float64)
        log of the state counts in each of the T thermodynamic states
    therm_energies : numpy.ndarray(shape=(T), dtype=numpy.float64)
        reduced free energies of the T thermodynamic states
    bias_energy_sequences : list of numpy.ndarray(shape=(X_i, T), dtype=numpy.float64)
        bias energies in the T thermodynamic states for all X samples
    scratch_T : numpy.ndarray(shape=(T), dtype=numpy.float64)
        scratch array
    new_therm_energies : numpy.ndarray(shape=(T), dtype=numpy.float64)
        target array for the reduced free energies of the T thermodynamic states
    """
    new_therm_energies[:] = _np.inf
    for i in range(len(bias_energy_sequences)):
        _mbar_update_therm_energies(
            <double*> _np.PyArray_DATA(log_therm_state_counts),
            <double*> _np.PyArray_DATA(therm_energies),
            <double*> _np.PyArray_DATA(bias_energy_sequences[i]),
            therm_energies.shape[0],
            bias_energy_sequences[i].shape[0],
            <double*> _np.PyArray_DATA(scratch_T),
            <double*> _np.PyArray_DATA(new_therm_energies))
    new_therm_energies -= new_therm_energies[0]

def get_conf_energies(
    _np.ndarray[double, ndim=1, mode="c"] log_therm_state_counts not None,
    _np.ndarray[double, ndim=1, mode="c"] therm_energies not None,
    bias_energy_sequences, # _np.ndarray[double, ndim=2, mode="c"]
    conf_state_sequences, # _np.ndarray[int, ndim=1, mode="c"]
    _np.ndarray[double, ndim=1, mode="c"] scratch_T not None,
    n_conf_states):
    r"""
    Calculate the reduced unbiased free energies conf_energies.
        
    Parameters
    ----------
    log_therm_state_counts : numpy.ndarray(shape=(T), dtype=numpy.float64)
        log of the state counts in each of the T thermodynamic states
    therm_energies : numpy.ndarray(shape=(T), dtype=numpy.float64)
        reduced free energies of the T thermodynamic states
    bias_energy_sequences : list of numpy.ndarray(shape=(X_i, T), dtype=numpy.float64)
        bias energies in the T thermodynamic states for all X samples
    conf_state_sequences : list of numpy.ndarray(shape=(X_i), dtype=numpy.intc)
        discrete states indices for all X samples
    scratch_T : numpy.ndarray(shape=(T), dtype=numpy.float64)
        scratch array
    n_conf_states : int
        number of discrete states (M)

    Returns
    -------
    conf_energies : numpy.ndarray(shape=(M), dtype=numpy.float64)
        reduced unbiased free energies
    biased_conf_energies : numpy.ndarray(shape=(T, M), dtype=numpy.float64)
        reduced bias energies in the T thermodynamic and M discrete states
    """
    conf_energies = _np.zeros(shape=(n_conf_states,), dtype=_np.float64)
    biased_conf_energies = _np.zeros(
        shape=(therm_energies.shape[0], n_conf_states), dtype=_np.float64)
    conf_energies[:] = _np.inf
    biased_conf_energies[:] = _np.inf
    for i in range(len(bias_energy_sequences)):
        _mbar_get_conf_energies(
            <double*> _np.PyArray_DATA(log_therm_state_counts),
            <double*> _np.PyArray_DATA(therm_energies),
            <double*> _np.PyArray_DATA(bias_energy_sequences[i]),
            <int*> _np.PyArray_DATA(conf_state_sequences[i]),
            therm_energies.shape[0],
            n_conf_states,
            conf_state_sequences[i].shape[0],
            <double*> _np.PyArray_DATA(scratch_T),
            <double*> _np.PyArray_DATA(conf_energies),
            <double*> _np.PyArray_DATA(biased_conf_energies))
    return conf_energies, biased_conf_energies

def normalize(
    _np.ndarray[double, ndim=1, mode="c"] scratch_M not None,
    _np.ndarray[double, ndim=1, mode="c"] therm_energies not None,
    _np.ndarray[double, ndim=1, mode="c"] conf_energies not None,
    _np.ndarray[double, ndim=2, mode="c"] biased_conf_energies not None):
    r"""
    Shift the reduced thermodynamic free energies therm_energies such that the
    unbiased thermodynamic free energy is zero.
        
    Parameters
    ----------
    scratch_M : numpy.ndarray(shape=(M), dtype=numpy.float64)
        scratch array
    therm_energies : numpy.ndarray(shape=(T), dtype=numpy.float64)
        reduced free energies of the T thermodynamic states
    conf_energies : numpy.ndarray(shape=(M), dtype=numpy.float64)
        reduced unbiased discrete state (cluster) free energies
    biased_conf_energies : numpy.ndarray(shape=(T, M), dtype=numpy.float64)
        reduced discrete state free energies for all combinations of
        T thermodynamic states and M discrete states
    """
    _mbar_normalize(
        therm_energies.shape[0],
        conf_energies.shape[0],
        <double*> _np.PyArray_DATA(scratch_M),
        <double*> _np.PyArray_DATA(therm_energies),
        <double*> _np.PyArray_DATA(conf_energies),
        <double*> _np.PyArray_DATA(biased_conf_energies))

def get_pointwise_unbiased_free_energies(
    k,
    _np.ndarray[double, ndim=1, mode="c"] log_therm_state_counts not None,
    bias_energy_sequences, # _np.ndarray[double, ndim=2, mode="c"]
    _np.ndarray[double, ndim=1, mode="c"] therm_energies not None,
    _np.ndarray[double, ndim=1, mode="c"] scratch_T,
    pointwise_unbiased_free_energies): # _np.ndarray[double, ndim=1, mode="c"]
    r'''
    Compute the pointwise free energies :math:`\mu^{k}(x)` for all x.

    Parameters
    ----------
    k : int or None
        thermodynamic state, if k is None, compute pointwise free energies
        of the unbiased ensemble.
    log_therm_state_counts : numpy.ndarray(shape=(T), dtype=numpy.float64)
        log of the state counts in each of the T thermodynamic states
    bias_energy_sequences : list of numpy.ndarray(shape=(X_i, T), dtype=numpy.float64)
        bias energies in the T thermodynamic states for all X samples
    therm_energies : numpy.ndarray(shape=(T), dtype=numpy.float64)
        reduced free energies of the T thermodynamic states
    scratch_T : numpy.ndarray(shape=(T), dtype=numpy.float64)
        scratch array for logsumexp operations
    pointwise_unbiased_free_energies : list of numpy.ndarray(shape=(X_i), dtype=numpy.float64)
        target arrays for the pointwise free energies
    '''

    if scratch_T is None:
        scratch_T = _np.zeros(log_therm_state_counts.shape[0], dtype=_np.float64)
    if k is None:
        k = -1
    assert len(bias_energy_sequences)==len(pointwise_unbiased_free_energies)
    for b, p in zip(bias_energy_sequences, pointwise_unbiased_free_energies):
        assert b.ndim == 2
        assert b.dtype == _np.float64
        assert p.ndim == 1
        assert p.dtype == _np.float64
        assert b.shape[0] == p.shape[0]
        assert b.shape[1] == log_therm_state_counts.shape[0]
        assert b.flags.c_contiguous
        assert p.flags.c_contiguous
    for i in range(len(bias_energy_sequences)):
        _mbar_get_pointwise_unbiased_free_energies(
            k,
            <double*> _np.PyArray_DATA(log_therm_state_counts),
            <double*> _np.PyArray_DATA(therm_energies),
            <double*> _np.PyArray_DATA(bias_energy_sequences[i]),
            log_therm_state_counts.shape[0],
            bias_energy_sequences[i].shape[0],
            <double*> _np.PyArray_DATA(scratch_T),
            <double*> _np.PyArray_DATA(pointwise_unbiased_free_energies[i]))

def estimate_therm_energies(
    therm_state_counts, bias_energy_sequences,
    maxiter=1000, maxerr=1.0E-8, therm_energies=None,
    n_conf_states=None, save_convergence_info=0, callback=None):
    r"""
    Estimate the thermodynamic free energies.
        
    Parameters
    ----------
    therm_state_counts : numpy.ndarray(shape=(T), dtype=numpy.intc)
        numbers of samples in the T thermodynamic states
    bias_energy_sequences : list of numpy.ndarray(shape=(X_i, T), dtype=numpy.float64)
        reduced bias energies in the T thermodynamic states for all X samples
    maxiter : int
        maximum number of iterations
    maxerr : float
        convergence criterion based on absolute change in free energies
    therm_energies : numpy.ndarray(shape=(T), dtype=numpy.float64), OPTIONAL
        initial guess for the reduced free energies of the T thermodynamic states
    n_conf_states : int, optional, default=None
        the number of configurational states in `conf_state_sequence`.
        If None, this is set to max(conf_state_sequence)+1.
    save_convergence_info : int, optional
        every save_convergence_info iteration steps, store the actual increment

    Returns
    -------
    therm_energies : numpy.ndarray(shape=(T), dtype=numpy.float64)
        reduced free energies of the T thermodynamic states
    increments : numpy.ndarray(dtype=numpy.float64, ndim=1)
        stored sequence of increments
    """
    T = therm_state_counts.shape[0]
    log_therm_state_counts = _np.log(therm_state_counts)
    if therm_energies is None:
        therm_energies = _np.zeros(shape=(T,), dtype=_np.float64)
    old_therm_energies = therm_energies.copy()
    increments = []
    sci_count = 0
    scratch = _np.zeros(shape=(T,), dtype=_np.float64)
    for m in range(maxiter):
        sci_count += 1
        update_therm_energies(
            log_therm_state_counts, old_therm_energies, bias_energy_sequences,
            scratch, therm_energies)
        delta_therm_energies = _np.abs(therm_energies - old_therm_energies)
        err = _np.max(delta_therm_energies)
        if sci_count == save_convergence_info:
            sci_count = 0
            increments.append(err)
        if callback is not None:
            try:
                callback(therm_energies=therm_energies,
                         old_therm_energies=old_therm_energies,
                         delta_therm_energies=delta_therm_energies,
                         iteration_step=m,
                         err=err,
                         maxerr=maxerr,
                         maxiter=maxiter)
            except CallbackInterrupt:
                break
        if err < maxerr:
            break
        else:
            old_therm_energies[:] = therm_energies[:]
    if err >= maxerr:
        _warn("MBAR did not converge: last increment = %.5e" % err, _NotConvergedWarning)
    if save_convergence_info == 0:
        increments = None
    else:
        increments = _np.array(increments, dtype=_np.float64)
    return therm_energies, increments

def estimate(
    therm_state_counts, bias_energy_sequences, conf_state_sequences,
    maxiter=1000, maxerr=1.0E-8, therm_energies=None,
    n_conf_states=None, save_convergence_info=0, callback=None):
    r"""
    Estimate the (un)biased reduced free energies and thermodynamic free energies.
        
    Parameters
    ----------
    therm_state_counts : numpy.ndarray(shape=(T), dtype=numpy.intc)
        numbers of samples in the T thermodynamic states
    bias_energy_sequences : list of numpy.ndarray(shape=(X_i, T), dtype=numpy.float64)
        reduced bias energies in the T thermodynamic states for all X samples
    conf_state_sequences : list of numpy.ndarray(shape=(X_i), dtype=numpy.float64)
        discrete state indices (cluster indices) for all X samples
    maxiter : int
        maximum number of iterations
    maxerr : float
        convergence criterion based on absolute change in free energies
    therm_energies : numpy.ndarray(shape=(T), dtype=numpy.float64), OPTIONAL
        initial guess for the reduced free energies of the T thermodynamic states
    n_conf_states : int, optional, default=None
        the number of configurational states in `conf_state_sequence`.
        If None, this is set to max(conf_state_sequence)+1.
    save_convergence_info : int, optional
        every save_convergence_info iteration steps, store the actual increment

    Returns
    -------
    therm_energies : numpy.ndarray(shape=(T), dtype=numpy.float64)
        reduced free energies of the T thermodynamic states
    conf_energies : numpy.ndarray(shape=(M), dtype=numpy.float64)
        reduced unbiased discrete state (cluster) free energies
    biased_conf_energies : numpy.ndarray(shape=(T, M), dtype=numpy.float64)
        reduced discrete state free energies for all combinations of
        T thermodynamic states and M discrete states
    increments : numpy.ndarray(dtype=numpy.float64, ndim=1)
        stored sequence of increments
    """
    T = therm_state_counts.shape[0]
    if n_conf_states is None:
        M = 1 + max([_np.max(s) for s in conf_state_sequences])
    else:
        M = n_conf_states
    assert len(conf_state_sequences)==len(bias_energy_sequences)
    for s, b in zip(conf_state_sequences, bias_energy_sequences):
        assert s.ndim == 1
        assert s.dtype == _np.intc
        assert b.ndim == 2
        assert b.dtype == _np.float64
        assert s.shape[0] == b.shape[0]
        assert b.shape[1] == T
        assert s.flags.c_contiguous
        assert b.flags.c_contiguous
    increments = []
    scratch_M = _np.zeros(shape=(M,), dtype=_np.float64)
    scratch_T = _np.zeros(shape=(T,), dtype=_np.float64)
    therm_energies, increments = estimate_therm_energies(
        therm_state_counts, bias_energy_sequences,
        maxiter=maxiter, maxerr=maxerr, therm_energies=therm_energies,
        save_convergence_info=save_convergence_info, callback=callback)
    conf_energies, biased_conf_energies = get_conf_energies(
        _np.log(therm_state_counts), therm_energies, bias_energy_sequences, conf_state_sequences,
        scratch_T, M)
    normalize(scratch_M, therm_energies, conf_energies, biased_conf_energies)
    return therm_energies, conf_energies, biased_conf_energies, increments



# This file is part of thermotools.
#
# Copyright 2015, 2016 Computational Molecular Biology Group, Freie Universitaet Berlin (GER)
#
# thermotools is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import numpy as np
import msmtools
import thermotools.tram as tram
import thermotools.tram_direct as tram_direct
import thermotools.trammbar as trammbar
import thermotools.trammbar_direct as trammbar_direct
import sys
from numpy.testing import assert_allclose

def tower_sample(distribution):
    cdf = np.cumsum(distribution)
    rnd = np.random.rand() * cdf[-1]
    return np.searchsorted(cdf, rnd)

def T_matrix(energy):
    n = energy.shape[0]
    metropolis = energy[np.newaxis, :] - energy[:, np.newaxis]
    metropolis[(metropolis < 0.0)] = 0.0
    selection = np.zeros((n, n))
    selection += np.diag(np.ones(n - 1) * 0.5, k=1)
    selection += np.diag(np.ones(n - 1) * 0.5, k=-1)
    selection[0, 0] = 0.5
    selection[-1, -1] = 0.5
    metr_hast = selection * np.exp(-metropolis)
    for i in range(metr_hast.shape[0]):
        metr_hast[i, i] = 0.0
        metr_hast[i, i] = 1.0 - metr_hast[i, :].sum()
    return metr_hast

def draw_transition_counts(transition_matrices, n_samples, x0):
    """generates a discrete Markov chain"""
    count_matrices = np.zeros(shape=transition_matrices.shape, dtype=np.intc)
    conf_state_sequence = np.zeros(
        shape=(transition_matrices.shape[0] * (n_samples + 1),), dtype=np.intc)
    state_counts = np.zeros(shape=transition_matrices.shape[0:2], dtype=np.intc)
    h = 0
    for K in range(transition_matrices.shape[0]):
        x = x0
        state_counts[K, x] += 1
        conf_state_sequence[h] = x
        h += 1
        for s in range(n_samples):
            x_new = tower_sample(transition_matrices[K, x, :])
            count_matrices[K, x, x_new] += 1
            x = x_new
            state_counts[K, x] += 1
            conf_state_sequence[h] = x
            h += 1
    return count_matrices, conf_state_sequence, state_counts

class TestRandom(object):
    @classmethod
    def setup_class(cls):
        n_therm_states = 4
        n_conf_states = 4
        n_samples = 10000
        cls.bias_energies = np.zeros(shape=(n_therm_states, n_conf_states), dtype=np.float64)
        cls.T = np.zeros(shape=(n_therm_states, n_conf_states, n_conf_states), dtype=np.float64)
        while True:
            # generate random stionary distributions
            for k in range(n_therm_states):
                cls.bias_energies[k, :] = -np.log(np.random.rand(n_conf_states))
                if k > 0:
                    cls.bias_energies[k, :] += np.random.rand()
            # generate transition matrices
            for k in range(n_therm_states):
                cls.T[k, :, :] = T_matrix(cls.bias_energies[k, :])
            cls.count_matrices, cls.conf_state_sequence, cls.state_counts = draw_transition_counts(
                cls.T, n_samples, 0)
            if msmtools.analysis.is_connected(cls.count_matrices.sum(axis=0), directed=True):
                break
        cls.bias_energies_sh = cls.bias_energies - cls.bias_energies[0, :] # hide pi^{0}_i
        cls.bias_energies_sh = np.ascontiguousarray(cls.bias_energies_sh[:, cls.conf_state_sequence])
        cls.n_therm_states = n_therm_states
        cls.n_conf_states = n_conf_states
        cls.n_samples = n_samples
    @classmethod
    def teardown_class(cls):
        pass
    def setup(self):
        pass
    def teardown(self):
        pass
    def test_tram(self):
        self.helper_tram(False, 0, False)
    def test_tram_direct(self):
        self.helper_tram(True, 0, False)
    def test_tram_direct_with_dTRAM_acceleration(self):
        self.helper_tram(True, 1, False)
    def test_trammbar_as_tram(self):
        self.helper_tram(False, 0, True)
    def test_trammbar_direct_as_tram(self):
        self.helper_tram(True, 0, True)
    def helper_tram(self, direct_space, N_dtram_accelerations, use_trammbar):
        if direct_space:
            _tram = tram_direct
        else:
            _tram = tram
        if use_trammbar:
            if direct_space:
                _tram = trammbar_direct
            else:
                _tram = trammbar
        ca = np.ascontiguousarray
        biased_conf_energies, conf_energies, therm_energies, log_lagrangian_mult, error_history, logL_history = _tram.estimate(
            self.count_matrices, self.state_counts,
            [ca(self.bias_energies_sh[:, 0:self.n_samples//2].T), ca(self.bias_energies_sh[:, self.n_samples//2:].T)],
            [self.conf_state_sequence[0:self.n_samples//2], self.conf_state_sequence[self.n_samples//2:]],
            maxiter=1000000, maxerr=1.0E-10, save_convergence_info=10, N_dtram_accelerations=N_dtram_accelerations)
        transition_matrices = _tram.estimate_transition_matrices(
            log_lagrangian_mult, biased_conf_energies, self.count_matrices, None)
        # check expectations (do a trivial test: recompute conf_energies with different functions)
        mu = np.zeros(shape=self.conf_state_sequence.shape[0], dtype=np.float64)
        _tram.get_pointwise_unbiased_free_energies(None, log_lagrangian_mult, biased_conf_energies,
            therm_energies, self.count_matrices,
            [ca(self.bias_energies_sh[:, 0:self.n_samples//2].T), ca(self.bias_energies_sh[:, self.n_samples//2:].T)],
            [self.conf_state_sequence[0:self.n_samples//2], self.conf_state_sequence[self.n_samples//2:]],
            self.state_counts, None, None,
            [mu[0:self.n_samples//2], mu[self.n_samples//2:]])
        counts,_ = np.histogram(
            self.conf_state_sequence, weights=np.exp(-mu), bins=self.n_conf_states)
        pmf = -np.log(counts)
        assert_allclose(pmf, conf_energies)
        biased_conf_energies -= np.min(biased_conf_energies)
        bias_energies =  self.bias_energies - np.min(self.bias_energies)
        nz = np.where(self.state_counts > 0)
        assert not np.any(np.isinf(log_lagrangian_mult[nz]))
        assert_allclose(biased_conf_energies, bias_energies, atol=0.1)
        assert_allclose(transition_matrices, self.T, atol=0.1)
        assert np.all(logL_history[-1] + 1.E-5 >= np.array(logL_history[0:-1]))
        # check exact identities of TRAM
        # (1) sum_j v_j T_ji + v_i = sum_j c_ij + sum_j c_ji
        for k in range(self.n_therm_states):
            lagrangian_mult = np.exp(log_lagrangian_mult[k, :])
            assert_allclose(
                lagrangian_mult.T.dot(transition_matrices[k, :, :]) + lagrangian_mult,
                self.count_matrices[k, :, :].sum(axis=0) + self.count_matrices[k, :, :].sum(axis=1))
        # (2) sum_jk v^k_j T^k_ji = sum_jk c^k_ji
        total = np.zeros(self.n_conf_states)
        for k in range(self.n_therm_states):
            lagrangian_mult = np.exp(log_lagrangian_mult[k, :])
            total += lagrangian_mult.T.dot(transition_matrices[k, :, :])
        assert_allclose(total, self.count_matrices.sum(axis=0).sum(axis=0))



# This file is part of thermotools.
#
# Copyright 2016 Computational Molecular Biology Group, Freie Universitaet Berlin (GER)
#
# thermotools is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import numpy as np
import thermotools.cset as cset
import thermotools.util as util

class TestCset(object):
    @classmethod
    def setup_class(cls):
        m1 = 10
        s1 = 1
        m2 = 20
        s2 = 1
        n_samples = 1000
        def u(x):
            return np.array([0.5 * ((x - m1) / s1)**2, 0.5 * ((x - m2) / s2)**2]).T
        x1 = np.random.randn(n_samples) * s1 + m1
        ttraj1 = np.zeros(n_samples, dtype=np.intc)
        dtraj1 = (x1 < 20).astype(np.intc)
        bias_traj1 = u(x1)
        x2 = np.random.randn(n_samples) * s2 + m2
        ttraj2 = np.ones(n_samples, dtype=np.intc)
        dtraj2 = (x2 < 20).astype(np.intc)
        bias_traj2 = u(x2)
        ttrajs = [ttraj1, ttraj2]
        dtrajs = [dtraj1, dtraj2]
        bias_trajs = [bias_traj1, bias_traj2]
        state_counts = util.state_counts(ttrajs, dtrajs)
        count_matrices = util.count_matrices(
            ttrajs, dtrajs, 1, sliding='sliding', sparse_return=False, nstates=2)
        cls.ttrajs = ttrajs
        cls.dtrajs = dtrajs
        cls.bias_trajs = bias_trajs
        cls.state_counts = state_counts
        cls.count_matrices = count_matrices
    @classmethod
    def teardown_class(cls):
        pass
    def setup(self):
        pass
    def teardown(self):
        pass
    def test_summed_count_matrix(self):
        csets, projected_cset = cset.compute_csets_TRAM(
            'summed_count_matrix', self.state_counts, self.count_matrices,
            ttrajs=self.ttrajs, dtrajs=self.dtrajs, bias_trajs=self.bias_trajs)
        np.testing.assert_allclose(csets[0], np.array([1]))
        np.testing.assert_allclose(csets[1], np.array([0, 1]))
        np.testing.assert_allclose(projected_cset, np.array([0,1]))
    # def test_strong_in_every_ensemble(self):
    #     csets, projected_cset = cset.compute_csets_TRAM(
    #         'strong_in_every_ensemble', self.state_counts, self.count_matrices, self.tram_sequence)
    #     np.testing.assert_allclose(csets[0], np.array([1]))
    #     np.testing.assert_allclose(csets[1], np.array([0, 1]))
    #     np.testing.assert_allclose(projected_cset, np.array([0,1]))
    def test_cset_neighbors(self):
        csets, projected_cset = cset.compute_csets_TRAM(
            'neighbors', self.state_counts, self.count_matrices, ttrajs=self.ttrajs,
            dtrajs=self.dtrajs, bias_trajs=self.bias_trajs, nn=1)
        np.testing.assert_allclose(csets[0], np.array([1]))
        np.testing.assert_allclose(csets[1], np.array([0, 1]))
        np.testing.assert_allclose(projected_cset, np.array([0,1]))
    def test_cset_post_hoc_RE(self):
        csets, projected_cset = cset.compute_csets_TRAM(
            'post_hoc_RE', self.state_counts, self.count_matrices, ttrajs=self.ttrajs,
            dtrajs=self.dtrajs, bias_trajs=self.bias_trajs)
        np.testing.assert_allclose(csets[0], np.array([]))
        np.testing.assert_allclose(csets[1], np.array([0, 1]))
        np.testing.assert_allclose(projected_cset, np.array([0, 1]))
    def test_cset_BAR_variance(self):
        csets, projected_cset = cset.compute_csets_TRAM(
            'BAR_variance', self.state_counts, self.count_matrices, ttrajs=self.ttrajs,
            dtrajs=self.dtrajs, bias_trajs=self.bias_trajs)
        np.testing.assert_allclose(csets[0], np.array([]))
        np.testing.assert_allclose(csets[1], np.array([0, 1]))
        np.testing.assert_allclose(projected_cset, np.array([0, 1]))
    def test_restrict(self):
        csets, projected_cset = cset.compute_csets_TRAM(
            'summed_count_matrix', self.state_counts, self.count_matrices,
            ttrajs=self.ttrajs, dtrajs=self.dtrajs, bias_trajs=self.bias_trajs)
        new_state_counts, new_count_matrices, new_dtrajs, new_bias_trajs = cset.restrict_to_csets(
            csets, self.state_counts, self.count_matrices, self.ttrajs,
            self.dtrajs, self.bias_trajs)
        np.testing.assert_allclose(new_count_matrices, self.count_matrices)
        np.testing.assert_allclose(new_state_counts, self.state_counts)
        for x,y in zip(self.bias_trajs, new_bias_trajs):
            np.testing.assert_allclose(x, y)
        for x,y in zip(self.dtrajs, new_dtrajs):
            np.testing.assert_allclose(x, y)
        csets, projected_cset = cset.compute_csets_TRAM(
            'post_hoc_RE', self.state_counts, self.count_matrices, ttrajs=self.ttrajs,
            dtrajs=self.dtrajs, bias_trajs=self.bias_trajs)
        new_state_counts, new_count_matrices, new_dtrajs, new_bias_trajs = cset.restrict_to_csets(
            csets, self.state_counts, self.count_matrices, self.ttrajs,
            self.dtrajs, self.bias_trajs)
        np.testing.assert_allclose(new_state_counts[0,:], 0)
        np.testing.assert_allclose(new_state_counts[1,:], self.state_counts[1,:])
        np.testing.assert_allclose(new_count_matrices[0,:,:], 0)
        np.testing.assert_allclose(new_count_matrices[1,:,:], self.count_matrices[1,:,:])
        assert len(new_bias_trajs[0])==0
        np.testing.assert_allclose(new_bias_trajs[1], self.bias_trajs[1])



# This file is part of thermotools.
#
# Copyright 2015 Computational Molecular Biology Group, Freie Universitaet Berlin (GER)
#
# thermotools is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import thermotools.dtram as dtram
import numpy as np
from numpy.testing import assert_allclose

def test_prior():
    assert_allclose(np.log(dtram.get_prior()), dtram.get_log_prior(), atol=1.0E-16)

def test_lognu_zero_counts():
    nm = 200
    nt = 100
    log_lagrangian_mult = np.zeros(shape=(nt, nm), dtype=np.float64)
    bias_energies = np.zeros(shape=(nt, nm), dtype=np.float64)
    conf_energies = np.zeros(shape=(nm,), dtype=np.float64)
    C_K_ij = np.zeros(shape=(nt, nm, nm), dtype=np.intc) # C_K_ii may have an internal prior
    scratch_i = np.zeros(shape=(nm,), dtype=np.float64)
    new_log_lagrangian_mult = np.zeros(shape=(nt, nm), dtype=np.float64)
    ref_log_lagrangian_mult = np.log(dtram.get_prior() * np.ones(shape=(nt, nm), dtype=np.float64))
    dtram.update_log_lagrangian_mult(
        log_lagrangian_mult, bias_energies, conf_energies, C_K_ij,
        scratch_i, new_log_lagrangian_mult)
    assert_allclose(new_log_lagrangian_mult, ref_log_lagrangian_mult, atol=1.0E-16)

def test_lognu_all_factors_unity():
    nm = 200
    nt = 100
    log_lagrangian_mult = np.zeros(shape=(nt, nm), dtype=np.float64)
    bias_energies = np.zeros(shape=(nt, nm), dtype=np.float64)
    conf_energies = np.zeros(shape=(nm,), dtype=np.float64)
    C_K_ij = np.ones(shape=(nt, nm, nm), dtype=np.intc)
    scratch_i = np.zeros(shape=(nm,), dtype=np.float64)
    new_log_lagrangian_mult = np.zeros(shape=(nt, nm), dtype=np.float64)
    ref_log_lagrangian_mult = np.log(nm*np.ones(shape=(nt, nm), dtype=np.float64))
    dtram.update_log_lagrangian_mult(
        log_lagrangian_mult, bias_energies, conf_energies, C_K_ij,
        scratch_i, new_log_lagrangian_mult)
    assert_allclose(new_log_lagrangian_mult, ref_log_lagrangian_mult, atol=1.0E-16)

def test_lognu_K_range():
    nm = 200
    nt = 100
    log_lagrangian_mult = np.zeros(shape=(nt, nm), dtype=np.float64)
    for K in range(nt):
        log_lagrangian_mult[K, :] = np.log(K + 1.0)
    bias_energies = np.zeros(shape=(nt, nm), dtype=np.float64)
    conf_energies = np.zeros(shape=(nm,), dtype=np.float64)
    C_K_ij = np.ones(shape=(nt, nm, nm), dtype=np.intc)
    scratch_i = np.zeros(shape=(nm,), dtype=np.float64)
    new_log_lagrangian_mult = np.zeros(shape=(nt, nm), dtype=np.float64)
    ref_log_lagrangian_mult = np.log(nm*np.ones(shape=(nt, nm), dtype=np.float64))
    dtram.update_log_lagrangian_mult(
        log_lagrangian_mult, bias_energies, conf_energies, C_K_ij,
        scratch_i, new_log_lagrangian_mult)
    assert_allclose(new_log_lagrangian_mult, ref_log_lagrangian_mult, atol=1.0E-16)


def test_fi_zero_counts():
    nm = 200
    nt = 100
    log_lagrangian_mult = np.zeros(shape=(nt, nm), dtype=np.float64)
    bias_energies = np.zeros(shape=(nt, nm), dtype=np.float64)
    conf_energies = np.zeros(shape=(nm,), dtype=np.float64)
    C_K_ij = np.zeros(shape=(nt, nm, nm), dtype=np.intc)
    scratch_TM = np.zeros(shape=(nt, nm), dtype=np.float64)
    scratch_M = np.zeros(shape=(nm,), dtype=np.float64)
    new_conf_energies = np.zeros(shape=(nm,), dtype=np.float64)
    dtram.update_conf_energies(
        log_lagrangian_mult, bias_energies, conf_energies, C_K_ij,
        scratch_TM, new_conf_energies)
    assert_allclose(new_conf_energies, 0.0, atol=1.0E-16)

def test_fi_all_factors_unity():
    nm = 200
    nt = 100
    log_lagrangian_mult = np.zeros(shape=(nt, nm), dtype=np.float64)
    bias_energies = np.zeros(shape=(nt, nm), dtype=np.float64)
    conf_energies = np.zeros(shape=(nm,), dtype=np.float64)
    C_K_ij = np.ones(shape=(nt, nm, nm), dtype=np.intc)
    scratch_TM = np.zeros(shape=(nt, nm), dtype=np.float64)
    scratch_M = np.zeros(shape=(nm,), dtype=np.float64)
    new_conf_energies = np.zeros(shape=(nm,), dtype=np.float64)
    dtram.update_conf_energies(
        log_lagrangian_mult, bias_energies, conf_energies, C_K_ij,
        scratch_TM, new_conf_energies)
    assert_allclose(new_conf_energies, 0.0, atol=1.0E-16)


def test_pij_zero_counts():
    nm = 200
    nt = 100
    log_lagrangian_mult = np.zeros(shape=(nt, nm), dtype=np.float64)
    bias_energies = np.zeros(shape=(nt, nm), dtype=np.float64)
    conf_energies = np.zeros(shape=(nm,), dtype=np.float64)
    C_K_ij = np.zeros(shape=(nt, nm, nm), dtype=np.intc)
    scratch_M = np.zeros(shape=(nm,), dtype=np.float64)
    p_K_ij = dtram.estimate_transition_matrices(
        log_lagrangian_mult, bias_energies, conf_energies, C_K_ij, scratch_M)
    ref_p_ij = np.eye(nm, dtype=np.float64)
    for K in range(nt):
        assert_allclose(p_K_ij[K, :, :], ref_p_ij, atol=1.0E-16)

def test_pij_all_factors_unity():
    nm = 200
    nt = 100
    log_lagrangian_mult = np.zeros(shape=(nt, nm), dtype=np.float64)
    bias_energies = np.zeros(shape=(nt, nm), dtype=np.float64)
    conf_energies = np.zeros(shape=(nm,), dtype=np.float64)
    C_K_ij = np.ones(shape=(nt, nm, nm), dtype=np.intc)
    scratch_M = np.zeros(shape=(nm,), dtype=np.float64)
    p_K_ij = dtram.estimate_transition_matrices(
        log_lagrangian_mult, bias_energies, conf_energies, C_K_ij, scratch_M)
    ref_p_ij = np.ones(shape=(nm, nm), dtype=np.float64) + \
        np.eye(nm, dtype=np.float64) * dtram.get_prior()
    ref_p_ij /= ref_p_ij.sum(axis=1)[:, np.newaxis]
    for K in range(nt):
        assert_allclose(p_K_ij[K, :, :], ref_p_ij, atol=1.0E-16)



# This file is part of thermotools.
#
# Copyright 2015, 2016 Computational Molecular Biology Group, Freie Universitaet Berlin (GER)
#
# thermotools is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

r"""
Python interface to the TRAM estimator's lowlevel functions.
"""

import numpy as _np
cimport numpy as _np

from warnings import warn as _warn
from msmtools.util.exceptions import NotConvergedWarning as _NotConvergedWarning

from .callback import CallbackInterrupt

__all__ = [
    'init_lagrangian_mult',
    'update_lagrangian_mult',
    'update_biased_conf_energies',
    'get_conf_energies',
    'normalize',
    'get_pointwise_unbiased_free_energies',
    'estimate_transition_matrix',
    'estimate_transition_matrices',
    'estimate']

DEF TRAMMBAR = True

cdef extern from "../tram/_tram.h":
    void _tram_init_lagrangian_mult(
        int *count_matrices, int n_therm_states, int n_conf_states, double *log_lagrangian_mult)
    void _tram_update_lagrangian_mult(
        double *log_lagrangian_mult, double *biased_conf_energies,
        int *count_matrices,  int* state_counts,
        int n_therm_states, int n_conf_states, double *scratch_M, double *new_log_lagrangian_mult)
    double _tram_update_biased_conf_energies(
        double *bias_energy_sequence, int *state_sequence, int seq_length, double *log_R_K_i,
        int n_therm_states, int n_conf_states, double *scratch_T,
        double *new_biased_conf_energies, int return_log_L)
    void _tram_get_conf_energies(
        double *bias_energy_sequence, int *state_sequence, int seq_length, double *log_R_K_i,
        int n_therm_states, int n_conf_states, double *scratch_T, double *conf_energies)
    void _tram_get_therm_energies(
        double *biased_conf_energies, int n_therm_states, int n_conf_states,
        double *scratch_M, double *therm_energies)
    void _tram_normalize(
        double *conf_energies, double *biased_conf_energies, double *therm_energies,
        int n_therm_states, int n_conf_states, double *scratch_M)
    void _tram_estimate_transition_matrix(
        double *log_lagrangian_mult, double *conf_energies, int *count_matrix,
        int n_conf_states, double *scratch_M, double *transition_matrix)
    double _tram_discrete_log_likelihood_lower_bound(
        double *log_lagrangian_mult, double *biased_conf_energies,
        int *count_matrices,  int *state_counts, int n_therm_states, int n_conf_states,
        double *scratch_M, double *scratch_MM,
        # TRAMMBAR below
        double *therm_energies, int *equilibrium_therm_state_counts,
        double overcounting_factor)
    void _tram_get_log_Ref_K_i(
        double *log_lagrangian_mult, double *biased_conf_energies, int *count_matrices,
        int *state_counts, int n_therm_states, int n_conf_states, double *scratch_M,
        double *log_R_K_i,
        # TRAMMBAR below
        double *therm_energies, int *equilibrium_therm_state_counts,
        double overcounting_factor)
    void _tram_get_pointwise_unbiased_free_energies(
        int k, double *bias_energy_sequence, double *therm_energies, int *state_sequence,
        int seq_length, double *log_R_K_i, int n_therm_states, int n_conf_states,
        double *scratch_T, double *pointwise_unbiased_free_energies)

def init_lagrangian_mult(
    _np.ndarray[int, ndim=3, mode="c"] count_matrices not None,
    _np.ndarray[double, ndim=2, mode="c"] log_lagrangian_mult not None):
    r"""
    Set the logarithm of the Lagrangian multipliers with an initial guess based
    on the transition counts

    Parameters
    ----------
    count_matrices : numpy.ndarray(shape=(T, M, M), dtype=numpy.intc)
        multistate count matrix
    log_lagrangian_mult : numpy.ndarray(shape=(T, M), dtype=numpy.float64)
        log of the Lagrangian multipliers (allocated but unset)
    """
    _tram_init_lagrangian_mult(
        <int*> _np.PyArray_DATA(count_matrices),
        log_lagrangian_mult.shape[0],
        log_lagrangian_mult.shape[1],
        <double*> _np.PyArray_DATA(log_lagrangian_mult))

def update_lagrangian_mult(
    _np.ndarray[double, ndim=2, mode="c"] log_lagrangian_mult not None,
    _np.ndarray[double, ndim=2, mode="c"] biased_conf_energies not None,
    _np.ndarray[int, ndim=3, mode="c"] count_matrices not None,
    _np.ndarray[int, ndim=2, mode="c"] state_counts not None,
    _np.ndarray[double, ndim=1, mode="c"] scratch_M not None,
    _np.ndarray[double, ndim=2, mode="c"] new_log_lagrangian_mult not None):
    r"""
    Update the logarithms of the Lagrangian multipliers

    Parameters
    ----------
    log_lagrangian_mult : numpy.ndarray(shape=(T, M), dtype=numpy.float64)
        log of the Lagrangian multipliers
    biased_conf_energies : numpy.ndarray(shape=(T, M), dtype=numpy.float64)
        reduced free energies
    count_matrices : numpy.ndarray(shape=(T, M, M), dtype=numpy.intc)
        multistate count matrix
    scratch_M : numpy.ndarray(shape=(M), dtype=numpy.float64)
        scratch array for logsumexp operations
    new_log_lagrangian_mult : numpy.ndarray(shape=(T, M), dtype=numpy.float64)
        target array for the log of the Lagrangian multipliers
    """
    _tram_update_lagrangian_mult(
        <double*> _np.PyArray_DATA(log_lagrangian_mult),
        <double*> _np.PyArray_DATA(biased_conf_energies),
        <int*> _np.PyArray_DATA(count_matrices),
        <int*> _np.PyArray_DATA(state_counts),
        log_lagrangian_mult.shape[0],
        log_lagrangian_mult.shape[1],
        <double*> _np.PyArray_DATA(scratch_M),
        <double*> _np.PyArray_DATA(new_log_lagrangian_mult))

def update_biased_conf_energies(
    _np.ndarray[double, ndim=2, mode="c"] log_lagrangian_mult not None,
    _np.ndarray[double, ndim=2, mode="c"] biased_conf_energies not None,
    _np.ndarray[int, ndim=3, mode="c"] count_matrices not None,
    bias_energy_sequences,
    state_sequences,
    _np.ndarray[int, ndim=2, mode="c"] state_counts not None,
    _np.ndarray[double, ndim=2, mode="c"] log_R_K_i not None,
    _np.ndarray[double, ndim=1, mode="c"] scratch_M not None,
    _np.ndarray[double, ndim=1, mode="c"] scratch_T not None,
    _np.ndarray[double, ndim=2, mode="c"] new_biased_conf_energies not None,
    _np.ndarray[double, ndim=2, mode="c"] scratch_MM,
    return_log_L=False,
    # TRAMMBAR below
    _np.ndarray[double, ndim=1, mode="c"] therm_energies=None,
    equilibrium_bias_energy_sequences=None,
    equilibrium_state_sequences=None,
    _np.ndarray[int, ndim=1, mode="c"] equilibrium_therm_state_counts=None,
    double overcounting_factor=1.0):
    r"""
    Update the reduced unbiased free energies

    Parameters
    ----------
    log_lagrangian_mult : numpy.ndarray(shape=(T, M), dtype=numpy.float64)
        log of the Lagrangian multipliers
    biased_conf_energies : numpy.ndarray(shape=(T, M), dtype=numpy.float64)
        reduced free energies
    count_matrices : numpy.ndarray(shape=(T, M, M), dtype=numpy.intc)
        multistate count matrix
    bias_energy_sequences : list of numpy.ndarray(shape=(X_i, T), dtype=numpy.float64)
        reduced bias energies in the T thermodynamic states for all X samples
    state_sequences : list of numpy.ndarray(shape=(X_i,), dtype=numpy.intc)
        Markov state indices for all X samples
    state_counts : numpy.ndarray(shape=(T, M), dtype=numpy.intc)
        number of visits to thermodynamic state K and Markov state i
    log_R_K_i : numpy.ndarray(shape=(T, M), dtype=numpy.float64)
        scratch array for sum of TRAM log pseudo-counts and biased_conf_energies
    scratch_M : numpy.ndarray(shape=(M), dtype=numpy.float64)
        scratch array for logsumexp operations
    scratch_T : numpy.ndarray(shape=(T), dtype=numpy.float64)
        scratch array for logsumexp operations
    new_biased_conf_energies : numpy.ndarray(shape=(T, M), dtype=numpy.float64)
        target array for the reduced free energies
    scratch_MM : numpy.ndarray(shape=(M, M), dtype=numpy.float64), optional
        scratch array for likelihood computation (only needed when
        return_log_L = True)
    return_log_L : bool
        If true, retrun the TRAM-log-likelihood.
    therm_energies : numpy.ndarray(shape=(T), dtype=numpy.float64)
        reduced thermodynamic free energies, must match `biased_conf_energies`
    equilibrium_bias_energy_sequences : list of numpy.ndarray(shape=(X_i, T), dtype=numpy.float64), optional
        reduced bias energies in the T thermodynamic states for all X equilibrium samples
    equilibrium_state_sequences : list of numpy.ndarray(shape=(X_i,), dtype=numpy.intc), optional
        Markov state indices for all X equilibrium samples
    equilibrium_therm_state_counts : numpy.ndarray(shape=(T), dtype=numpy.intc)
        number of equilibrium frames per thermodynamic state, can be zero
    overcounting_factor : double, default = 1.0
        Sets the relative statistical weight of equilibrium and non-equilibrium
        frames. An overcounting_factor of value n means that every
        non-equilibrium frame is assumed to be repeated n times in the data.
    """
    new_biased_conf_energies[:] = _np.inf
    get_log_Ref_K_i(log_lagrangian_mult, biased_conf_energies,
                    count_matrices, state_counts, scratch_M, log_R_K_i,
                    therm_energies=therm_energies,
                    equilibrium_therm_state_counts=equilibrium_therm_state_counts,
                    overcounting_factor=overcounting_factor)
    log_L = 0.0
    for i in range(len(bias_energy_sequences)):
        log_L += _tram_update_biased_conf_energies(
            <double*> _np.PyArray_DATA(bias_energy_sequences[i]),
            <int*> _np.PyArray_DATA(state_sequences[i]),
            state_sequences[i].shape[0],
            <double*> _np.PyArray_DATA(log_R_K_i),
            log_lagrangian_mult.shape[0],
            log_lagrangian_mult.shape[1],
            <double*> _np.PyArray_DATA(scratch_T),
            <double*> _np.PyArray_DATA(new_biased_conf_energies),
            int(return_log_L))
    if TRAMMBAR:
        if equilibrium_bias_energy_sequences is not None:
            log_L *= overcounting_factor
            new_biased_conf_energies -= _np.log(overcounting_factor)
            for i in range(len(equilibrium_bias_energy_sequences)):
                log_L += _tram_update_biased_conf_energies(
                    <double*> _np.PyArray_DATA(equilibrium_bias_energy_sequences[i]),
                    <int*> _np.PyArray_DATA(equilibrium_state_sequences[i]),
                    equilibrium_state_sequences[i].shape[0],
                    <double*> _np.PyArray_DATA(log_R_K_i),
                    log_lagrangian_mult.shape[0],
                    log_lagrangian_mult.shape[1],
                    <double*> _np.PyArray_DATA(scratch_T),
                    <double*> _np.PyArray_DATA(new_biased_conf_energies),
                    int(return_log_L))
    else:
        assert equilibrium_bias_energy_sequences is None
        assert equilibrium_state_sequences is None
        assert equilibrium_therm_state_counts is None
    if return_log_L:
        assert scratch_MM is not None
        log_L += _tram_discrete_log_likelihood_lower_bound(
            <double*> _np.PyArray_DATA(log_lagrangian_mult),
            <double*> _np.PyArray_DATA(new_biased_conf_energies),
            <int*> _np.PyArray_DATA(count_matrices),
            <int*> _np.PyArray_DATA(state_counts),
            state_counts.shape[0],
            state_counts.shape[1],
            <double*> _np.PyArray_DATA(scratch_M),
            <double*> _np.PyArray_DATA(scratch_MM),
            <double*> _np.PyArray_DATA(therm_energies) if therm_energies is not None else NULL,
            <int*> _np.PyArray_DATA(equilibrium_therm_state_counts) if equilibrium_therm_state_counts is not None else NULL,
            overcounting_factor)
        return log_L

def get_log_Ref_K_i(
    _np.ndarray[double, ndim=2, mode="c"] log_lagrangian_mult not None,
    _np.ndarray[double, ndim=2, mode="c"] biased_conf_energies not None,
    _np.ndarray[int, ndim=3, mode="c"] count_matrices not None,
    _np.ndarray[int, ndim=2, mode="c"] state_counts not None,
    _np.ndarray[double, ndim=1, mode="c"] scratch_M not None,
    _np.ndarray[double, ndim=2, mode="c"] log_R_K_i not None,
    # TRAMMBAR below
    _np.ndarray[double, ndim=1, mode="c"] therm_energies=None,
    _np.ndarray[int, ndim=1, mode="c"] equilibrium_therm_state_counts=None,
    double overcounting_factor=1.0):
    r"""
    Computes the sum of TRAM log pseudo-counts and biased_conf_energies.

    Parameters
    ----------
    log_lagrangian_mult : numpy.ndarray(shape=(T, M), dtype=numpy.float64)
        log of the Lagrangian multipliers
    biased_conf_energies : numpy.ndarray(shape=(T, M), dtype=numpy.float64)
        reduced free energies
    count_matrices : numpy.ndarray(shape=(T, M, M), dtype=numpy.intc)
        multistate count matrix
    state_counts : numpy.ndarray(shape=(T, M), dtype=numpy.intc)
        number of visits to thermodynamic state K and Markov state i
    scratch_M : numpy.ndarray(shape=(M), dtype=numpy.float64)
        scratch array for logsumexp operations
    log_R_K_i : numpy.ndarray(shape=(T, M), dtype=numpy.float64)
        target array for sum of TRAM log pseudo-counts and biased_conf_energies
    therm_energies : numpy.ndarray(shape=(T), dtype=numpy.float64)
        reduced thermodynamic free energies, must match `biased_conf_energies`
    equilibrium_therm_state_counts : numpy.ndarray(shape=(T), dtype=numpy.intc)
        number of equilibrium frames per thermodynamic state, can be zero
    overcounting_factor : double, default = 1.0
        Sets the relative statistical weight of equilibrium and non-equilibrium
        frames. An overcounting_factor of value n means that every
        non-equilibrium frame is assumed to be repeated n times in the data.
    """
    if not TRAMMBAR:
        assert equilibrium_therm_state_counts is None
    _tram_get_log_Ref_K_i(
        <double*> _np.PyArray_DATA(log_lagrangian_mult),
        <double*> _np.PyArray_DATA(biased_conf_energies),
        <int*> _np.PyArray_DATA(count_matrices),
        <int*> _np.PyArray_DATA(state_counts),
        log_lagrangian_mult.shape[0],
        log_lagrangian_mult.shape[1],
        <double*> _np.PyArray_DATA(scratch_M),
        <double*> _np.PyArray_DATA(log_R_K_i),
        <double*> _np.PyArray_DATA(therm_energies) if therm_energies is not None else NULL,
        <int*> _np.PyArray_DATA(equilibrium_therm_state_counts) if equilibrium_therm_state_counts is not None else NULL,
        overcounting_factor)

def get_conf_energies(
    bias_energy_sequences,
    state_sequences,
    _np.ndarray[double, ndim=2, mode="c"] log_R_K_i not None,
    _np.ndarray[double, ndim=1, mode="c"] scratch_T not None,
    # TRAMMBAR below
    equilibrium_bias_energy_sequences=None,
    equilibrium_state_sequences=None,
    double overcounting_factor=1.0):
    r"""
    Update the reduced unbiased free energies

    Parameters
    ----------
    bias_energy_sequences : list of numpy.ndarray(shape=(X_i, T), dtype=numpy.float64)
        reduced bias energies in the T thermodynamic states for all X samples
    state_sequence : list of numpy.ndarray(shape=(X_i,), dtype=numpy.intc)
        Markov state indices for all X samples
    log_R_K_i : numpy.ndarray(shape=(T, M), dtype=numpy.float64)
        precomputed sum of TRAM log pseudo-counts and biased_conf_energies
    scratch_T : numpy.ndarray(shape=(T), dtype=numpy.float64)
        scratch array for logsumexp operations
    equilibrium_bias_energy_sequences : list of numpy.ndarray(shape=(X_i, T), dtype=numpy.float64), optional
        reduced bias energies in the T thermodynamic states for all X equilibrium samples
    equilibrium_state_sequences : list of numpy.ndarray(shape=(X_i,), dtype=numpy.intc), optional
        Markov state indices for all X equilibrium samples
    overcounting_factor : double, default = 1.0
        Sets the relative statistical weight of equilibrium and non-equilibrium
        frames. An overcounting_factor of value n means that every
        non-equilibrium frame is assumed to be repeated n times in the data.

    Returns
    -------
    conf_energies : numpy.ndarray(shape=(M,), dtype=numpy.float64)
        unbiased (Markov) free energies
    """
    conf_energies = _np.zeros(shape=(log_R_K_i.shape[1],), dtype=_np.float64)
    conf_energies[:] = _np.inf
    for i in range(len(bias_energy_sequences)):
        _tram_get_conf_energies(
            <double*> _np.PyArray_DATA(bias_energy_sequences[i]),
            <int*> _np.PyArray_DATA(state_sequences[i]),
            state_sequences[i].shape[0],
            <double*> _np.PyArray_DATA(log_R_K_i),
            log_R_K_i.shape[0],
            log_R_K_i.shape[1],
            <double*> _np.PyArray_DATA(scratch_T),
            <double*> _np.PyArray_DATA(conf_energies))
    if TRAMMBAR:
        if equilibrium_bias_energy_sequences is not None:
            conf_energies -= _np.log(overcounting_factor)
            for i in range(len(equilibrium_bias_energy_sequences)):
                _tram_get_conf_energies(
                    <double*> _np.PyArray_DATA(equilibrium_bias_energy_sequences[i]),
                    <int*> _np.PyArray_DATA(equilibrium_state_sequences[i]),
                    equilibrium_state_sequences[i].shape[0],
                    <double*> _np.PyArray_DATA(log_R_K_i),
                    log_R_K_i.shape[0],
                    log_R_K_i.shape[1],
                    <double*> _np.PyArray_DATA(scratch_T),
                    <double*> _np.PyArray_DATA(conf_energies))
    else:
        assert equilibrium_bias_energy_sequences is None
        assert equilibrium_state_sequences is None
    return conf_energies

def get_therm_energies(
    _np.ndarray[double, ndim=2, mode="c"] biased_conf_energies not None,
    _np.ndarray[double, ndim=1, mode="c"] scratch_M not None):
    r"""
    Update the reduced unbiased free energies

    Parameters
    ----------
    biased_conf_energies : numpy.ndarray(shape=(T, M), dtype=numpy.float64)
        reduced bias energies in the T thermodynamic and M discrete states
    scratch_M : numpy.ndarray(shape=(M), dtype=numpy.float64)
        scratch array for logsumexp operations

    Returns
    -------
    therm_energies : numpy.ndarray(shape=(T), dtype=numpy.float64)
        reduced thermodynamic free energies
    """
    therm_energies = _np.zeros(shape=(biased_conf_energies.shape[0],), dtype=_np.float64)
    _tram_get_therm_energies(
        <double*> _np.PyArray_DATA(biased_conf_energies),
        biased_conf_energies.shape[0],
        biased_conf_energies.shape[1],
        <double*> _np.PyArray_DATA(scratch_M),
        <double*> _np.PyArray_DATA(therm_energies))
    return therm_energies

def normalize(
    _np.ndarray[double, ndim=1, mode="c"] conf_energies not None,
    _np.ndarray[double, ndim=2, mode="c"] biased_conf_energies not None,
    _np.ndarray[double, ndim=1, mode="c"] therm_energies not None,
    _np.ndarray[double, ndim=1, mode="c"] scratch_M not None):
    r"""
    Update the reduced unbiased free energies

    Parameters
    ----------
    conf_energies : numpy.ndarray(shape=(M), dtype=numpy.float64)
        unbiased reduced bias energies in the M discrete states
    biased_conf_energies : numpy.ndarray(shape=(T, M), dtype=numpy.float64)
        reduced bias energies in the T thermodynamic and M discrete states
    therm_energies : numpy.ndarray(shape=(T), dtype=numpy.float64)
        reduced thermodynamic free energies
    scratch_M : numpy.ndarray(shape=(M), dtype=numpy.float64)
        scratch array for logsumexp operations
    """
    _tram_normalize(
        <double*> _np.PyArray_DATA(conf_energies),
        <double*> _np.PyArray_DATA(biased_conf_energies),
        <double*> _np.PyArray_DATA(therm_energies),
        biased_conf_energies.shape[0],
        biased_conf_energies.shape[1],
        <double*> _np.PyArray_DATA(scratch_M))

def get_pointwise_unbiased_free_energies(
    k,
    _np.ndarray[double, ndim=2, mode="c"] log_lagrangian_mult not None,
    _np.ndarray[double, ndim=2, mode="c"] biased_conf_energies not None,
    _np.ndarray[double, ndim=1, mode="c"] therm_energies not None,
    _np.ndarray[int, ndim=3, mode="c"] count_matrices not None,
    bias_energy_sequences,
    state_sequences,
    _np.ndarray[int, ndim=2, mode="c"] state_counts not None,
    _np.ndarray[double, ndim=1, mode="c"] scratch_M,
    _np.ndarray[double, ndim=1, mode="c"] scratch_T,
    pointwise_unbiased_free_energies,
    # TRAMMBAR below
    _np.ndarray[int, ndim=1, mode="c"] equilibrium_therm_state_counts=None,
    double overcounting_factor=1.0):
    r'''
    Compute the pointwise free energies :math:`\mu^{k}(x)` for all x.

    Parameters
    ----------
    k : int or None
        thermodynamic state, if k is None, compute pointwise free energies
        of the unbiased ensemble.
    log_lagrangian_mult : numpy.ndarray(shape=(T, M), dtype=numpy.float64)
        log of the Lagrangian multipliers
    biased_conf_energies : numpy.ndarray(shape=(T, M), dtype=numpy.float64)
        reduced free energies
    therm_energies : numpy.ndarray(shape=(T), dtype=numpy.float64)
        reduced thermodynamic free energies
    count_matrices : numpy.ndarray(shape=(T, M, M), dtype=numpy.intc)
        multistate count matrix
    bias_energy_sequences : list of numpy.ndarray(shape=(X_i, T), dtype=numpy.float64)
        reduced bias energies in the T thermodynamic states for all X samples
    state_sequences : list of numpy.ndarray(shape=(X_i,), dtype=numpy.intc)
        Markov state indices for all X samples
    state_counts : numpy.ndarray(shape=(T, M), dtype=numpy.intc)
        number of visits to thermodynamic state K and Markov state i
    scratch_M : numpy.ndarray(shape=(M), dtype=numpy.float64)
        scratch array for logsumexp operations
    scratch_T : numpy.ndarray(shape=(T), dtype=numpy.float64)
        scratch array for logsumexp operations
    pointwise_unbiased_free_energies : list of numpy.ndarray(shape=(X_i), dtype=numpy.float64)
        target arrays for the pointwise free energies
    equilibrium_therm_state_counts : numpy.ndarray(shape=(T), dtype=numpy.intc), optional
        number of equilibrium frames per thermodynamic state, can be zero
    overcounting_factor : double, default = 1.0
        Sets the relative statistical weight of equilibrium and non-equilibrium
        frames. An overcounting_factor of value n means that every
        non-equilibrium frame is assumed to be repeated n times in the data.
    '''

    log_R_K_i = _np.zeros(shape=(state_counts.shape[0],state_counts.shape[1]), dtype=_np.float64)
    if scratch_T is None:
        scratch_T = _np.zeros(shape=(state_counts.shape[0]), dtype=_np.float64)
    if scratch_M is None:
        scratch_M = _np.zeros(shape=(state_counts.shape[1]), dtype=_np.float64)
    get_log_Ref_K_i(
        log_lagrangian_mult, biased_conf_energies,
        count_matrices, state_counts, scratch_M, log_R_K_i,
        therm_energies=therm_energies,
        equilibrium_therm_state_counts=equilibrium_therm_state_counts,
        overcounting_factor=overcounting_factor)
    if k is None:
        k = -1
    assert len(state_sequences) == len(bias_energy_sequences) == len(pointwise_unbiased_free_energies)
    for s, b, p in zip(state_sequences, bias_energy_sequences, pointwise_unbiased_free_energies):
        assert s.ndim == 1
        assert s.dtype == _np.intc
        assert b.ndim == 2
        assert b.dtype == _np.float64
        assert p.ndim == 1
        assert p.dtype == _np.float64
        assert s.shape[0] == b.shape[0] == p.shape[0]
        assert b.shape[1] == count_matrices.shape[0]
        assert s.flags.c_contiguous
        assert b.flags.c_contiguous
        assert p.flags.c_contiguous
    for i in range(len(bias_energy_sequences)):
        _tram_get_pointwise_unbiased_free_energies(
            k,
            <double*> _np.PyArray_DATA(bias_energy_sequences[i]),
            <double*> _np.PyArray_DATA(therm_energies),
            <int*> _np.PyArray_DATA(state_sequences[i]),
            state_sequences[i].shape[0], 
            <double*> _np.PyArray_DATA(log_R_K_i),
            log_R_K_i.shape[0],
            log_R_K_i.shape[1],
            <double*> _np.PyArray_DATA(scratch_T),
            <double*> _np.PyArray_DATA(pointwise_unbiased_free_energies[i]))

def estimate_transition_matrices(
    _np.ndarray[double, ndim=2, mode="c"] log_lagrangian_mult not None,
    _np.ndarray[double, ndim=2, mode="c"] biased_conf_energies not None,
    _np.ndarray[int, ndim=3, mode="c"] count_matrices not None,
    _np.ndarray[double, ndim=1, mode="c"] scratch_M):
    r"""
    Compute the transition matrices for all thermodynamic states

    Parameters
    ----------
    log_lagrangian_mult : numpy.ndarray(shape=(T, M), dtype=numpy.float64)
        log of the Lagrangian multipliers
    biased_conf_energies : numpy.ndarray(shape=(T, M), dtype=numpy.float64)
        reduced unbiased free energies
    count_matrices : numpy.ndarray(shape=(T, M, M), dtype=numpy.intc)
        multistate count matrix
    scratch_M : numpy.ndarray(shape=(M), dtype=numpy.float64)
        scratch array for logsumexp operations

    Returns
    -------
    p_K_ij : numpy.ndarray(shape=(T, M, M), dtype=numpy.float64)
        transition matrices for all thermodynamic states
    """
    if scratch_M is None:
        scratch_M = _np.zeros(shape=(count_matrices.shape[1],), dtype=_np.float64)
    p_K_ij = _np.zeros(
        shape=(count_matrices.shape[0], count_matrices.shape[1], count_matrices.shape[2]),
        dtype=_np.float64)
    for K in range(log_lagrangian_mult.shape[0]):
        p_K_ij[K, :, :] = estimate_transition_matrix(
            log_lagrangian_mult, biased_conf_energies, count_matrices, scratch_M, K)[:, :]
    return p_K_ij

def estimate_transition_matrix(
    _np.ndarray[double, ndim=2, mode="c"] log_lagrangian_mult not None,
    _np.ndarray[double, ndim=2, mode="c"] biased_conf_energies not None,
    _np.ndarray[int, ndim=3, mode="c"] count_matrices not None,
    _np.ndarray[double, ndim=1, mode="c"] scratch_M,
    therm_state):
    r"""
    Compute the transition matrices for all thermodynamic states

    Parameters
    ----------
    log_lagrangian_mult : numpy.ndarray(shape=(T, M), dtype=numpy.float64)
        log of the Lagrangian multipliers
    biased_conf_energies : numpy.ndarray(shape=(T, M), dtype=numpy.float64)
        reduced unbiased free energies
    count_matrices : numpy.ndarray(shape=(T, M, M), dtype=numpy.intc)
        multistate count matrix
    scratch_M : numpy.ndarray(shape=(M), dtype=numpy.float64)
        scratch array for logsumexp operations
    therm_state : int
        target thermodynamic state

    Returns
    -------
    transition_matrix : numpy.ndarray(shape=(M, M), dtype=numpy.float64)
        transition matrix for the target thermodynamic state
    """
    if scratch_M is None:
        scratch_M = _np.zeros(shape=(count_matrices.shape[1],), dtype=_np.float64)
    transition_matrix = _np.zeros(
        shape=(biased_conf_energies.shape[1], biased_conf_energies.shape[1]), dtype=_np.float64)
    _tram_estimate_transition_matrix(
        <double*> _np.PyArray_DATA(_np.ascontiguousarray(log_lagrangian_mult[therm_state, :])),
        <double*> _np.PyArray_DATA(_np.ascontiguousarray(biased_conf_energies[therm_state, :])),
        <int*> _np.PyArray_DATA(_np.ascontiguousarray(count_matrices[therm_state, :, :])),
        biased_conf_energies.shape[1],
        <double*> _np.PyArray_DATA(scratch_M),
        <double*> _np.PyArray_DATA(transition_matrix))
    return transition_matrix

def log_likelihood_lower_bound(
    _np.ndarray[double, ndim=2, mode="c"] log_lagrangian_mult not None,
    _np.ndarray[double, ndim=2, mode="c"] biased_conf_energies not None,
    _np.ndarray[int, ndim=3, mode="c"] count_matrices not None,
    bias_energy_sequences,
    state_sequences,
    _np.ndarray[int, ndim=2, mode="c"] state_counts not None,
    _np.ndarray[double, ndim=2, mode="c"] log_R_K_i,
    _np.ndarray[double, ndim=1, mode="c"] scratch_M,
    _np.ndarray[double, ndim=1, mode="c"] scratch_T,
    _np.ndarray[double, ndim=2, mode="c"] scratch_TM,
    _np.ndarray[double, ndim=2, mode="c"] scratch_MM,
    # TRAMMBAR below
    _np.ndarray[double, ndim=1, mode="c"] therm_energies=None,
    equilibrium_bias_energy_sequences=None,
    equilibrium_state_sequences=None,
    _np.ndarray[int, ndim=1, mode="c"] equilibrium_therm_state_counts=None,
    double overcounting_factor=1.0):
    r"""
    Computes a lower bound on the TRAM log-likelihood

    Parameters
    ----------
    log_lagrangian_mult : numpy.ndarray(shape=(T, M), dtype=numpy.float64)
        log of the Lagrangian multipliers
    biased_conf_energies : numpy.ndarray(shape=(T, M), dtype=numpy.float64)
        reduced free energies
    count_matrices : numpy.ndarray(shape=(T, M, M), dtype=numpy.intc)
        multistate count matrix
    bias_energy_sequences : list of numpy.ndarray(shape=(X_i, T), dtype=numpy.float64)
        reduced bias energies in the T thermodynamic states for all X samples
    state_sequences : list of numpy.ndarray(shape=(X_i,), dtype=numpy.intc)
        Markov state indices for all X samples
    state_counts : numpy.ndarray(shape=(T, M), dtype=numpy.intc)
        number of visits to thermodynamic state K and Markov state i
    log_R_K_i : numpy.ndarray(shape=(T, M), dtype=numpy.float64)
        scratch array for sum of TRAM log pseudo-counts and biased_conf_energies
    scratch_M : numpy.ndarray(shape=(M), dtype=numpy.float64)
        scratch array for logsumexp operations
    scratch_T : numpy.ndarray(shape=(T), dtype=numpy.float64)
        scratch array for logsumexp operations
    scratch_TM : numpy.ndarray(shape=(T, M), dtype=numpy.float64)
        scratch array for logsumexp operations
    scratch_MM : numpy.ndarray(shape=(M, M), dtype=numpy.float64)
        scratch array for likelihood computation
    therm_energies : numpy.ndarray(shape=(T), dtype=numpy.float64)
        reduced thermodynamic free energies, must match `biased_conf_energies`
    equilibrium_bias_energy_sequences : list of numpy.ndarray(shape=(X_i, T), dtype=numpy.float64), optional
        reduced bias energies in the T thermodynamic states for all X equilibrium samples
    equilibrium_state_sequences : list of numpy.ndarray(shape=(X_i,), dtype=numpy.intc), optional
        Markov state indices for all X equilibrium samples
    equilibrium_therm_state_counts : numpy.ndarray(shape=(T), dtype=numpy.intc)
        number of equilibrium frames per thermodynamic state, can be zero
    overcounting_factor : double, default = 1.0
        Sets the relative statistical weight of equilibrium and non-equilibrium
        frames. An overcounting_factor of value n means that every
        non-equilibrium frame is assumed to be repeated n times in the data.

    Note
    ----
    mu_i^{(k)} are regarded as functions of log_lagrangian_mult,
    biased_conf_energies, count_matrices, bias_energy_sequence,
    state_sequence and state_counts. By itself that doesn't have to be
    normalized. Internally exp(-new_biased_conf_energies) are computed
    such that they are the correct normalization constants for mu_i^{(k)}.

    The lower bound on the log-likelihood is computed by first projecting
    the transition matrix to close feasible point and then computing
    the (exact) log-likelihood for the combination of mu_i^{(k)},
    new_biased_conf_energies and this projected transition matrix.
    Because the projection of the transition matrix is not optimal
    (in the likelihood sense) this yields only a lower bound on the
    true log-likelihood.
    """
    T = biased_conf_energies.shape[0]
    M = biased_conf_energies.shape[1]
    if log_R_K_i is None:
        log_R_K_i = _np.zeros((T, M), dtype=_np.float64)
    if scratch_M is None:
        scratch_M = _np.zeros((M,), dtype=_np.float64)
    if scratch_T is None:
        scratch_T = _np.zeros((T,), dtype=_np.float64)
    if scratch_TM is None:
        scratch_TM = _np.zeros((T, M), dtype=_np.float64)
    if scratch_MM is None:
        scratch_MM = _np.zeros((M, M), dtype=_np.float64)
    return update_biased_conf_energies(
        log_lagrangian_mult, biased_conf_energies, count_matrices,
        bias_energy_sequences, state_sequences, state_counts,
        log_R_K_i, scratch_M, scratch_T, scratch_TM, scratch_MM, True,
        therm_energies=therm_energies,
        equilibrium_bias_energy_sequences=equilibrium_bias_energy_sequences,
        equilibrium_state_sequences=equilibrium_state_sequences,
        equilibrium_therm_state_counts=equilibrium_therm_state_counts,
        overcounting_factor=overcounting_factor)

def estimate(count_matrices, state_counts, bias_energy_sequences, state_sequences,
    maxiter=1000, maxerr=1.0E-8, save_convergence_info=0,
    biased_conf_energies=None, log_lagrangian_mult=None, callback=None, N_dtram_accelerations=0,
    equilibrium_therm_state_counts=None,
    equilibrium_bias_energy_sequences=None, equilibrium_state_sequences=None,
    overcounting_factor = 1.0):
    r"""
    Estimate the reduced discrete state free energies and thermodynamic free energies

    Parameters
    ----------
    count_matrices : numpy.ndarray(shape=(T, M, M), dtype=numpy.intc)
        transition count matrices for all T thermodynamic states
    state_counts : numpy.ndarray(shape=(T, M), dtype=numpy.intc)
        state counts for all M discrete and T thermodynamic states
    bias_energy_sequences : list of numpy.ndarray(shape=(X_i, T), dtype=numpy.float64)
        reduced bias energies in the T thermodynamic states for all X samples
    state_sequences : list of numpy.ndarray(shape=(X_i), dtype=numpy.float64)
        discrete state indices for all X samples
    maxiter : int
        maximum number of iterations
    maxerr : float
        convergence criterion based on absolute change in free energies
    save_convergence_info : int, optional
        every save_convergence_info iteration steps, store the actual increment
        and the actual loglikelihood
    biased_conf_energies : numpy.ndarray(shape=(T, M), dtype=numpy.float64), OPTIONAL
        initial guess for the reduced discrete state free energies for all T thermodynamic states
    log_lagrangian_mult : numpy.ndarray(shape=(T, M), dtype=numpy.float64), OPTIONAL
        initial guess for the logarithm of the Lagrangian multipliers
    N_dtram_accelerations : int
        not used
    equilibrium_therm_state_counts : numpy.ndarray(shape=(T,), dtype=numpy.intc), optional
        number of equilibrium frames per thermodynamic state, can be zero
    equilibrium_bias_energy_sequences : list of numpy.ndarray(shape=(X_i, T), dtype=numpy.float64), optional
        reduced bias energies in the T thermodynamic states for all X equilibrium samples
    equilibrium_state_sequences : list of numpy.ndarray(shape=(X_i), dtype=numpy.float64), optional
        discrete Markov state indices for all X equilibrium samples
    overcounting_factor : double, default = 1.0
        Sets the relative statistical weight of equilibrium and non-equilibrium
        frames. An overcounting_factor of value n means that every
        non-equilibrium frame is assumed to be repeated n times in the data.

    Returns
    -------
    biased_conf_energies : numpy.ndarray(shape=(T, M), dtype=numpy.float64)
        reduced discrete state free energies for all T thermodynamic states
    conf_energies : numpy.ndarray(shape=(M), dtype=numpy.float64)
        reduced unbiased discrete state free energies
    therm_energies : numpy.ndarray(shape=(M), dtype=numpy.float64)
        reduced thermodynamic free energies
    log_lagrangian_mult : numpy.ndarray(shape=(T, M), dtype=numpy.float64)
        logarithm of the Lagrangian multipliers
    increments : numpy.ndarray(dtype=numpy.float64, ndim=1)
        stored sequence of increments
    loglikelihoods : numpy.ndarray(dtype=numpy.float64, ndim=1)
        stored sequence of loglikelihoods

    Note
    ----
    The self-consitent iteration terminates when

    .. math::
       \max\{\max_{i,k}{\Delta \pi_i^k}, \max_k \Delta f^k \}<\mathrm{maxerr}.

    Different termination criteria can be implemented with the callback
    function. Raising `CallbackInterrupt` in the callback will cleanly
    terminate the iteration.
    """
    if biased_conf_energies is None:
        biased_conf_energies = _np.zeros(shape=state_counts.shape, dtype=_np.float64)
    if log_lagrangian_mult is None:
        log_lagrangian_mult = _np.zeros(shape=state_counts.shape, dtype=_np.float64)
        init_lagrangian_mult(count_matrices, log_lagrangian_mult)
    if not TRAMMBAR:
        assert equilibrium_therm_state_counts is None
    increments = []
    loglikelihoods = []
    sci_count = 0
    assert len(state_sequences) == len(bias_energy_sequences)
    for s, b in zip(state_sequences, bias_energy_sequences):
        assert s.ndim == 1
        assert s.dtype == _np.intc
        assert b.ndim == 2
        assert b.dtype == _np.float64
        assert s.shape[0] == b.shape[0]
        assert b.shape[1] == count_matrices.shape[0]
        assert s.flags.c_contiguous
        assert b.flags.c_contiguous
    if TRAMMBAR:
        if equilibrium_state_sequences is not None:
            assert len(equilibrium_state_sequences) == len(equilibrium_bias_energy_sequences)
            for s, b in zip(equilibrium_state_sequences, equilibrium_bias_energy_sequences):
                assert s.ndim == 1
                assert s.dtype == _np.intc
                assert b.ndim == 2
                assert b.dtype == _np.float64
                assert s.shape[0] == b.shape[0]
                assert b.shape[1] == count_matrices.shape[0]
                assert s.flags.c_contiguous
                assert b.flags.c_contiguous
    else:
        assert equilibrium_bias_energy_sequences is None
        assert equilibrium_state_sequences is None
    log_R_K_i = _np.zeros(shape=state_counts.shape, dtype=_np.float64)
    scratch_T = _np.zeros(shape=(count_matrices.shape[0],), dtype=_np.float64)
    scratch_M = _np.zeros(shape=(count_matrices.shape[1],), dtype=_np.float64)
    scratch_MM = _np.zeros(shape=count_matrices.shape[1:3], dtype=_np.float64)
    therm_energies = get_therm_energies(biased_conf_energies, scratch_M)
    old_biased_conf_energies = biased_conf_energies.copy()
    old_log_lagrangian_mult = log_lagrangian_mult.copy()
    old_stat_vectors = _np.zeros(shape=state_counts.shape, dtype=_np.float64)
    old_therm_energies = therm_energies.copy()
    for _m in range(maxiter):
        sci_count += 1 
        update_lagrangian_mult(
            old_log_lagrangian_mult, biased_conf_energies, count_matrices, state_counts,
            scratch_M, log_lagrangian_mult)
        l = update_biased_conf_energies(
            log_lagrangian_mult, old_biased_conf_energies, count_matrices,
            bias_energy_sequences, state_sequences, state_counts,
            log_R_K_i, scratch_M, scratch_T, biased_conf_energies,
            scratch_MM, sci_count == save_convergence_info,
            therm_energies=old_therm_energies,
            equilibrium_bias_energy_sequences=equilibrium_bias_energy_sequences,
            equilibrium_state_sequences=equilibrium_state_sequences,
            equilibrium_therm_state_counts=equilibrium_therm_state_counts,
            overcounting_factor=overcounting_factor)

        therm_energies = get_therm_energies(biased_conf_energies, scratch_M)
        stat_vectors = _np.exp(therm_energies[:, _np.newaxis] - biased_conf_energies)
        delta_therm_energies = _np.abs(therm_energies - old_therm_energies)
        delta_stat_vectors =  _np.abs(stat_vectors - old_stat_vectors)
        err = max(_np.max(delta_therm_energies), _np.max(delta_stat_vectors))
        if sci_count == save_convergence_info:
            sci_count = 0
            increments.append(err)
            loglikelihoods.append(l)
        if callback is not None:
            try:
                callback(biased_conf_energies=biased_conf_energies,
                         log_lagrangian_mult=log_lagrangian_mult,
                         therm_energies=therm_energies,
                         stat_vectors=stat_vectors,
                         old_biased_conf_energies=old_biased_conf_energies,
                         old_log_lagrangian_mult=old_log_lagrangian_mult,
                         old_stat_vectors=old_stat_vectors,
                         old_therm_energies=old_therm_energies,
                         iteration_step=_m,
                         err=err,
                         maxerr=maxerr,
                         maxiter=maxiter)
            except CallbackInterrupt:
                break
        if err < maxerr:
            break
        else:
            shift = _np.min(biased_conf_energies)
            biased_conf_energies -= shift
            old_biased_conf_energies[:] = biased_conf_energies
            old_log_lagrangian_mult[:] = log_lagrangian_mult[:]
            old_therm_energies[:] = therm_energies[:] - shift
            old_stat_vectors[:] = stat_vectors[:]
    conf_energies = get_conf_energies(bias_energy_sequences, state_sequences, log_R_K_i, scratch_T,
                         equilibrium_bias_energy_sequences=equilibrium_bias_energy_sequences,
                         equilibrium_state_sequences=equilibrium_state_sequences,
                         overcounting_factor=overcounting_factor)
    therm_energies = get_therm_energies(biased_conf_energies, scratch_M)
    normalize(conf_energies, biased_conf_energies, therm_energies, scratch_M)
    if err >= maxerr:
        _warn("TRAM did not converge: last increment = %.5e" % err, _NotConvergedWarning)
    if save_convergence_info == 0:
        increments = None
        loglikelihoods = None
    else:
        increments = _np.array(increments, dtype=_np.float64)
        loglikelihoods = _np.array(loglikelihoods, dtype=_np.float64)

    return biased_conf_energies, conf_energies, therm_energies, log_lagrangian_mult, \
        increments, loglikelihoods

def simple_error(callback=None):
    r"""
    Stopping condition for `estimate`. Can be given as the value of `callback`.
    
    Stop the estimation when the difference of the biased energies
    (logarithms of the joint probability of conformational state and
    thermodynamic state) between two iterations is smaller than `maxerr`.

    Parameters
    ----------
    callback : optional
        user call back. Because `simple_error` takes the `callback`
        slot of `estimate`, this allows to chain `simple_error` with
        another call back.
    """
    def function(**kwargs):
        biased_conf_energies = kwargs['biased_conf_energies']
        old_biased_conf_energies = kwargs['old_biased_conf_energies']
        maxerr = kwargs['maxerr']
        if callback is not None:
            callback(**kwargs)
        if _np.max(_np.abs(biased_conf_energies - old_biased_conf_energies)) < maxerr:
            raise CallbackInterrupt('biased configuration energies have converged')
    return function
